/* $Header: /usr/local/src/dhcpd/RCS/dhcp.c,v 1.22 1998/07/02 18:35:22 irwin Exp $ */

/***********************************************************
        Copyright 1996 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <net/if.h>
#include <net/route.h>
#include <netdb.h>
#include <syslog.h>
#include <errno.h>
#include <arpa/inet.h>  /* inet_ntoa */


#include "defs.h"
#include "bootp.h"
#include "bootpd.h"
#include "dhcpd.h"
#include "dhcp.h"
#include "dynamic.h"
#include "readfile.h"
#include "perm.h"
#include "getif.h"
#include "hwaddr.h"
#include "utils.h"


extern u_short bootpc_port; /* host byte order */
extern u_short bootps_port; /* host byte order */
extern int debug,s,pktlen;
extern boolean compact_messages;
hash_tbl *boundhshtbl;
hash_tbl *ipboundhshtbl;


struct unknownhost *unknown;
unsigned int num_unknown;

extern struct in_addr my_ip_addr;
static int i;
static unsigned int ui;
int init=0;
unsigned long dhcp_alarm;
unsigned int leases;
char buf[MAXCLIIDLEN];
static unsigned char send_buf[2000];
static struct bootp *rpkt;
u_char enet_all_zeros[];
u_char htype_ethernet;

extern struct sockaddr_in recv_addr;
void send_dhcp_reply __P((struct bootp *pkt, int32 send_len, unsigned int32 dest));
void send_dhcpnack __P((struct bootp *pkt, struct shared_bindata *cliid, char *msg, unsigned int max_len, boolean cliid_explicit));
void order_insert();
struct host *create_stub __P((struct host* hp, struct client_info *cl));
struct host *create_lease __P((struct host* hp, struct shared_bindata *cliid,
				struct shared_bindata *vend, 
				struct shared_bindata *user, unsigned secs));
int do_req_opts __P((struct host *hp, u_char *req_opts, u_char *bp, struct opts *cur, struct flag *sent_options));
int do_opts __P((struct host *hp, u_char *bp, struct opts *cur, struct flag *sent_options));

struct host *hp;
struct shared_bindata *cliid,*vend,*user;
unsigned secs;
unsigned int discovers_received,offers_sent;
unsigned int requests_received,acks_sent,nacks_sent;
unsigned int ips_expired,ips_reclaimed,declines_received;
unsigned int releases_received;
unsigned int informs_received;
unsigned long gcinterval;
unsigned int bootp_requests_received,bootp_replys_sent;
unsigned long expire_infinite_dynamics;

unsigned long now,starttime;
struct in_addr my_ip_addr;
int pktlen;
void report_cliid();
char * cliid_btoa();
extern char * hostname;

/* like ctime(), but takes a time_t instead of a time_t *.
   Don't call twice in same expr! */
char *cctime(at)
unsigned long at;
{
 char *r;

 r=ctime((time_t*)&at);
 r[strlen(r)-1]=0;
 return r;
}

void init_dhcp()
{
 struct rtentry rt;
 struct sockaddr_in addr;
 struct lastbind *binding;
 struct client_info cl;
 hash_datum *data;
 int len,i,rc;
 unsigned hashcode, hashDel_hashcode;
 struct host *hp, *hp2;
 struct in_addr tmpaddr, hashDel_iaddr;
 char flag;
 boolean force_garbage_collect=FALSE;
 char str1[2*MAXSTRINGLEN], str2[MAXSTRINGLEN];


 lastbindhshtbl=hash_Init(HASHTABLESIZE);
 boundhshtbl=hash_Init(HASHTABLESIZE);
 ipboundhshtbl=hash_Init(HASHTABLESIZE);
 cache2disk=hash_Init(HASHTABLESIZE);

 if(debug)
	report(LOG_INFO,"using last binds dir %s",bindingsdir);
 if((lastbindpermanent=hashf_Init(bindingsdir,HASHTABLESIZE)) == NULL){
	report(LOG_ERR,"Could not open last binds dir : %s",LASTBINDDIR);
	exit(1);
 }

 hash_Reset(boundhshtbl,free_host);
 hash_Reset(ipboundhshtbl,free_host);
 hash_Reset(lastbindhshtbl,free_lastbind);
 hash_Reset(cache2disk,free_lastbind);

 i=0; /* items inserted into lastbindhshtbl */
 leases=0;	 /* items inserted into the bound hash tables */
 unknown=NULL;
 now=time(0);
 tmpaddr.s_addr=0xffffffff;
 if (debug > 11) {
	report(LOG_INFO, " Entries in bindings directory follow...");
	report(LOG_INFO, " Each will be added to lastbindings structure.");
	report(LOG_INFO, " Unexpired/infinite dhcp leases will be reinstated as current bindings.");
	report(LOG_INFO, " BootP assignments for dynamic addresses will be reinstated as current bindings,");
	if (expire_infinite_dynamics)
		report(LOG_INFO, "    (but only if their last_touch is more recent than expire_infinite_dynamics (%d) seconds)", 
			expire_infinite_dynamics);
	else 
		report(LOG_INFO, "    regardless of their last_touch age (expire_infinite_dynamics is not set)");
 }
 for(data=hashf_FirstEntry(lastbindpermanent,&len);data != NULL;
			data=hashf_NextEntry(lastbindpermanent,&len)){
	binding=data2bind(data,len);
	if(!binding){
		report(LOG_ERR,"  data2bind() returned NULL, corrupt binding entry found on disk, skipping this entry");
		continue;
		}
	binding->linkcount++;

	if(debug>11){
		report_cliid(LOG_INFO,binding->cliid);
		/* must build message up, since cctime() shouldn't be called twice in same expr */
		sprintf(str1, "    IP=%s, expires=%s, code=%s, replies=%d, ", 
			inet_ntoa(binding->last_ip), 
			((binding->expire == INFINITY) ? "never" : cctime(binding->expire)),
			((binding->code < 6) ? codename[binding->code] : "UNKNOWN"),
			binding->replys);
		sprintf(str2, "last touched at %s", cctime(binding->last_touch));
		strcat(str1,str2);
		report(LOG_INFO, str1);

	}

#ifdef DEBUG
	if (debug > 15)
		report(LOG_DEBUG, "    init_dhcp(): will insert into lastbindhshtbl...");
#endif
	rc=hash_Insert(lastbindhshtbl, hash_HashFunction(binding->cliid->data,binding->cliid->length), lastbindcmp,binding,binding);
	if (rc) 
		report(LOG_ERR, "    init_dhcp(): error inserting binding in lastbindhshtbl!");

	i++;

	/* look up IP address in bootptab */
	hashcode=hash_HashFunction((u_char*)&(binding->last_ip.s_addr),4);
	hp=(struct host *)hash_Lookup(iphashtable,hashcode,ipcmp, &(binding->last_ip.s_addr));
	if(!hp) {
		if (debug > 11)
			report(LOG_INFO,"    IP address no longer in bootptab, will not create lease");
		continue;
	}

	if (binding->expire != INFINITY) { /* DHCP assignment, OR BootP dynamic that already expired due to expire_infinite_dynamics */
		if (now >= binding->expire) {
			if ((binding->code == DHCP_FLAG) || (binding->code==RESERVED_FLAG)) {
				if (debug > 11)
					report(LOG_INFO, "    Was for DHCP, but has expired; will not create lease");
				continue;
			} else if (binding->code == BOOTP_FLAG) { 
				if (debug > 11)
					report(LOG_INFO, "    Was for BootP dynamic, but previously expired due to expire_infinite_dynamics; will not create lease");
				continue;
			}
		} else { /* now < binding->expire */
			if (debug > 11) 
				report(LOG_INFO, "    Was for DHCP and has not expired; will create lease"); 
		}
	} else { /*BOOTP assignment */
		if (binding->code == RESERVED_FLAG) { /* static */
			if (debug > 11)
				report(LOG_INFO, "    Was for BOOTP static assignment, will not create lease");
			continue;
		} else if (binding->code == BOOTP_FLAG) { /* dynamic */
			if (expire_infinite_dynamics) {
				if (now >= binding->last_touch + expire_infinite_dynamics) {
					/* It was for a BOOTP dynamic, older than expire_infinite_dynamics.  So we don't want to create a lease.
					   Normally, garbage collection would have found this, and set expire to 'then', expired the lease,
					   and updated lastbinding.  But if the server went down between the time this should have happened
					   and the next garbage collection, or the server's been down for some time, or the expire_infinite_dynamics
					   has only recently been set (or reduced), we'll find entries like this in the lastbind directory.
					   We want to clean them up, by setting their expire value to 'now', so we won't have these in this
					   state forever.  But it is not safe to just do that now, since that could cause a write to this
					   directory, which we are in the middle of reading; we don't have any sort of locking.  So the 
					   solution is a kludge; we'll go ahead and let the binding be created, and allow regular garbage collection
					   to clean it up.  We'll speed that up by forcing garbage collection early. */
					if (debug > 11) {
						report(LOG_INFO, "    Was for BOOTP dynamic, but older than expire_infinite_dynamics...");
						report(LOG_INFO, "     don't want lease, but will create one to allow garbage_collect to clean up expire time");
					}
					force_garbage_collect=TRUE;
				} else {
					if (debug > 11)
						report(LOG_INFO, "    Was for BOOTP dynamic and newer than expire_infinite_dynamics; will create lease");
				}
			} else {
				if (debug > 11)
					report(LOG_INFO, "    Was for BOOTP dynamic, will create lease");
			}
		} else {
			report(LOG_ERR, "    init_dhcp(): invalid binding->code");
			continue;
		}
	} /* DHCP or BOOTP assignment */


	bzero(&cl,sizeof(cl));
	cl.cliid=binding->cliid;
	cl.vend=binding->vend;
	cl.user=binding->user;
	cl.clhashcode=hash_HashFunction(binding->cliid->data,binding->cliid->length);

	/* create a struct host, which we *may* use later; if we choose not to, we'll free it */
	if (!(hp=create_stub(hp,&cl))) {
		report(LOG_ERR,"    create_stub() returned NULL; Unable to create stub for client, skipping!");
		continue; /* walk lastbindpermanent */
	}
	if(!create_lease(hp,cl.cliid,cl.vend,cl.user,0)){
		hp->linkcount=1;  /* to force free_host() to free it, as it starts =0 */
		free_host(hp);
		report(LOG_ERR,"    create_lease() returned NULL; Unable to create lease for client, skipping!");
		continue; /* walk lastbindpermanent */
	}

	hp->linkcount=1;	/* it started =0 */

	hp->expire=binding->expire;
	if (hp->expire != INFINITY)
		hp->flags.expire=1;	/* else = 0 */

	hp->code=binding->code;

	/* if IP address is dynamic, make sure access policy still allows this client. */
	if (hp->flags.dynamic) {

		/* build 'flags' to indicate the protocol and privileges associated with the client */
		if (binding->code==DHCP_FLAG || binding->code==RESERVED_FLAG)
			flag = DHCP_ACCESS;
		else if (binding->code==BOOTP_FLAG)
			flag = BOOTP_ACCESS;
		else {
			report(LOG_ERR, "   init_dhcp(): invalid binding->code");
			free_host(hp);
			continue; /* walk lastbindpermanent */
		}
		/* look for clientid in bootptab */
		hp2 =(struct host *)hash_Lookup(clhashtable, cl.clhashcode,cliidcmp,cl.cliid);
		/* if not found, should we look again using haddr as key?  But how, since binding only
		   records cliid, not haddr. */
		if (hp2) {
			flag |=  REG_ACCESS;
			if (hp2->flags.roaming)
				flag |= ROAMING_ACCESS;
		}

		/* here's where we actually check that the client satisfies the access policy.
		   We are calling use_addr() instead of access_network() for two reasons.  First,
		   acces_network() requires that we pass the network's restriction code; we could 
		   look that up, but use_addr() will do that for us before it calls access_network().
		   Second, when use_addr() finds that the client meets the access policies, it will
		   take care of removing the IP address from the pool of available dynamics.    
		   Note that because use_addr() normally checks the client's IP address against a giaddr,
		   and we don't care about that here, we tell it to skip that test by calling with force=1. */
		if (!use_addr(hp, tmpaddr, flag, 1)){
			if (debug > 11)
				report(LOG_INFO,"     address is dynamic, but client (now) denied by access policy; will not create lease");
			free_host(hp);
			continue; /* walk lastbindpermanent */
		}
	}

#ifdef DEBUG
	if (debug > 15)
		report(LOG_DEBUG, "    Will insert into ipboundhshtbl...");
#endif
	if(hash_Insert(ipboundhshtbl,hashcode,ipcmp,(hash_datum *)&(hp->iaddr.s_addr),(hash_datum *)hp)){
		report(LOG_ERR,"    Found unexpired lease, but insert into ipboundhshtbl failed (cliid=%s)", cliid_btoa(hp->cl));
		report(LOG_ERR,"    Probably duplicate entry for %s (bindings database corrupt)",inet_ntoa(hp->iaddr));
		report(LOG_ERR,"    could not reinstate lease");
		free_host(hp); /* since hp->linkcount == 1, this will truly free it */
		continue; /* walk lastbindpermanent */
	}
	/* note that hp->linkcount still == 1, now reflecting reality (it is in ipboundhshtbl) */

	hp->linkcount++; /* to reflect the insert we're about to perform */
#ifdef DEBUG
	if (debug > 15)
		report(LOG_DEBUG, "    Will insert into boundhshtbl...");
#endif
	rc=hash_Insert(boundhshtbl,cl.clhashcode,boundcliidcmp,hp->cl,hp);
	if (rc)  {
		report(LOG_ERR, "    error inserting binding into boundhshtbl (cliid=%s)", cliid_btoa(hp->cl));
		report(LOG_ERR, "    could not reinstate lease for %s", inet_ntoa(hp->iaddr));
		report(LOG_ERR, "    will delete binding just added to ipboundhshtbl (ip=%s)", inet_ntoa(hp->iaddr));
		free_host(hp); /* doesn't really free it since linkcount == 2 */
		/* We already inserted it into ipboundhshtbl.  Since we failed to insert it in boundhshtbl, let's
		   remove it from ipboundhshtbl to keep the two in sync. */
		/* Make standalone copy of hashcode and key to pass to hash_Delete(), since hash_Delete() may free hp (thereby
		   freeing hp->iphashcode and hp->iaddr), while hash_Delete() still needs to access those to continue walking the
		   hashtable to find any more members that match. */
		hashDel_hashcode = hp->iphashcode;
		hashDel_iaddr.s_addr = hp->iaddr.s_addr;
		if (hash_Delete(ipboundhshtbl, hashDel_hashcode, ipcmp, (hash_datum *)&(hashDel_iaddr.s_addr), free_host)) {
			report(LOG_ERR, "    error deleting binding from ipboundhshtbl (ip=%s)", inet_ntoa(hp->iaddr));
			/* at this point, things are probably seriously screwed up */
		}
		hp->linkcount--;	/* to 1 */
		free_host(hp);	/* really frees it */
		continue;
	} else {
		if(debug>11)
			report(LOG_INFO,"    Reinstated lease for %s, expires %s",
				inet_ntoa(hp->iaddr),
				(binding->expire==INFINITY) ? "never" : cctime(binding->expire));
		leases++;
#ifdef DEBUG
		if (debug > 15)
			report(LOG_DEBUG, "    incremented leases counter to %d", leases);
#endif
	}
 } /* walk lastbindpermanent */

 if(lastbindpermanent->next){
	fclose(lastbindpermanent->next);
	lastbindpermanent->next=NULL;
 }
 if(debug)
 	report(LOG_INFO,"%d lastbindings (%d current bindings) were retrieved from disk",i, leases);
 if(max_cachequeue==0)
 	max_cachequeue=20;
 if(flush_interval==0)
 	flush_interval=30*60L;
 if(gcinterval==0)
	gcinterval=30*60L;
 dhcp_alarm=time(0)+gcinterval;
 flush_alarm=time(0)+flush_interval;
 num_unknown=0;
 informs_received=0;
 discovers_received=0;
 requests_received=0;
 releases_received=0;
 declines_received=0;
 ips_reclaimed=0;
 ips_expired=0;
 acks_sent=0;
 nacks_sent=0;
 offers_sent=0;

 if (force_garbage_collect) {
	if (debug > 11)
		report(LOG_INFO, "   Forcing garbage collection to force expiration of BootP dynamics older than expire_infinite_dynamics");
	dhcp_garbage_collect();
 }

}

/* if this is a dhcp packet, return the type otherwise 0
   if the options in this packet are poorly formed, return -1 
 */
char dhcp_msg_type(pkt)
struct bootp *pkt;
{
 u_char *cp;
 unsigned hashcode;


 if(bcmp(pkt->bp_vend,vm_rfc1048,4))
	return 0;
 for(cp=(u_char*)pkt,i=4+((u_long)pkt->bp_vend-(u_long)pkt); i<pktlen ;){
	if(cp[i]==TAG_PAD){
		i++;
		continue;
	}
	if(cp[i]==TAG_END)
			return 0;

	if(cp[i]==TAG_DHCP_MSG){
		i+=2;
		if(i>=pktlen || cp[i]>DHCPINFORM || cp[i]<DHCPDISCOVER)
			return -1;
		return cp[i];
	}
		
	i++;
	if(i==pktlen || ((int)(cp[i]+i)>=pktlen))
		return -1;
	i+=cp[i]+1;
  }
  return 0;
}


char * cliid_btoa(cl)
struct shared_bindata *cl;
/* return string representation of cliid 'cl' */
/* note use of static buffer; don't call more than once from single expr */
{
	boolean printable = TRUE; /* indicates the entire cliid is printable characters */
	static char str[MAXCLIIDLEN];
	unsigned int i;

	if (!cl)
		return "";

	/* set 'printable' */
	for (i=0; i<cl->length && printable; i++)
		if (!(ispunct(cl->data[i])||isalnum(cl->data[i])||isspace(cl->data[i])))
		 printable = FALSE;

	if (printable) {
		bcopy(cl->data,str,cl->length);
		str[cl->length]=0;
	} else
		for (i=0, str[0]=0; i<cl->length; i++)
			sprintf(str+strlen(str),"%02X ",cl->data[i]);
		
	return str;
}

void report_cliid(log,cl)
int log;
struct shared_bindata *cl;
/* print the cliid in 'cl' at log level 'log' */
{
	if (!cl)
		report(LOG_ERR, "report_cliid(): cl is NULL");
	else
		report(log, "  Client id: %s len=%d", cliid_btoa(cl), cl->length);
}


/* Read some info from pkt, filling out various fields in cl. 
   Return 0 on success, 1 on error */
/* On error, note we don't free cl->cliidhw we may malloc - whoever calls us has the responsibility of
   freeing the entire struct cl eventually, which takes care of cl->cliidhw too */
int dhcp_opts(pkt,cl)
struct bootp *pkt;
struct client_info *cl;
{
	u_char *cp,overload;
	int i;

	if (pkt->bp_htype >= hwinfocnt) {
		/* if htype unknown, ignore packet */
		if (debug > 2)
			report(LOG_NOTICE, "  unknown hardware type: %d, ignoring bootp client (hlen=%d, chaddr=%s)", pkt->bp_htype, pkt->bp_hlen, haddrtoa(pkt->bp_chaddr,BP_CHADDR_LEN));
		return 1;
	}


	overload=0;
	cl->cliidhw=(struct shared_bindata*) smalloc(sizeof(struct shared_bindata)+pkt->bp_hlen);
	cl->cliidhw->length=haddrlength(pkt->bp_htype)+1;
	if (cl->cliidhw->length-1 != pkt->bp_hlen) {
		if (compact_messages && debug)
			report(LOG_INFO, "DHCPpkterror %s [%s]: bad address length %d != %d",
				haddrtoa2(pkt->bp_chaddr, pkt->bp_hlen), inet_ntoa(pkt->bp_giaddr), pkt->bp_hlen, cl->cliidhw->length - 1);
		if (debug > 2)
			report(LOG_NOTICE, "  bad addr len from from %s address %s", netname(pkt->bp_htype),
				haddrtoa(pkt->bp_chaddr,cl->cliidhw->length-1 ));
		return 1;
	}

	if (pkt->bp_htype == HTYPE_ETHERNET) {
		/* rule common bogus Ethernet addresses of all zeros */
		if (!bcmp((char *) &pkt->bp_chaddr, (char *) &enet_all_zeros, haddrlength(HTYPE_ETHERNET))) {
			if (debug > 2)
				report(LOG_NOTICE, "  bad Ethernet chaddr (all zeros), ignoring DHCP client");
			return 1;
		}
		/* rule out broadcast and multicast Ethernet addreses, which are bogus client enet addresses */
		if (pkt->bp_chaddr[0] % 2 == 1) {
			if (debug > 2)
				report(LOG_NOTICE, "  bad Ethernet chaddr %s (broadcast or multicast), ignoring DHCP client", haddrtoa(pkt->bp_chaddr, pkt->bp_hlen));
			return 1;
		}
	}

	cl->cliidhw->data[0]=pkt->bp_htype;
	cl->cliidhw->linkcount=1;
	bcopy(pkt->bp_chaddr,cl->cliidhw->data+1,cl->cliidhw->length-1); 


	for(i=VEND_OFFSET,cp=(u_char*)pkt;i<pktlen;){
		switch(cp[i]){
		case TAG_PAD:i++;continue;
		case TAG_DHCP_OVL:
			overload=cp[i+2]&3;
		case TAG_END:
			if(overload&1){  /* file overload */
				i=FILE_OFFSET;
				overload &= ~1; /* Turn off bit 1 -- mkl */
				continue;
			}
			if(overload&2){  /* sname overload */
				i=SNAME_OFFSET;
				overload &= ~2; /* Turn off bit 2 -- mkl */
				continue;
			}
			if(!cl->cliid){
				cl->cliid=cl->cliidhw;
				cl->cliid->linkcount++;
				/* cl->cliid_explicit remains FALSE */
			}
			return 0;
		case TAG_MESSAGE:
			cl->msg=(char*)smalloc(cp[i+1]+1);
			bcopy(cp+i+2,cl->msg,cp[i+1]);
			cl->msg[cp[i+1]+1]=0;
			break;
		case TAG_DHCP_USERCLASS:
			cl->user=(struct shared_bindata*) smalloc(sizeof(struct shared_bindata)+cp[i+1]-1);
			bcopy(&cp[i+2],cl->user->data,cp[i+1]);
			cl->user->linkcount=1;
			cl->user->length=cp[i+1];
			break;
		case TAG_VENDOR:
			cl->vend=(struct shared_bindata*) smalloc(sizeof(struct shared_bindata)+cp[i+1]-1);
			bcopy(&cp[i+2],cl->vend->data,cp[i+1]);
			cl->vend->linkcount=1;
			cl->vend->length=cp[i+1];
			break;
		case TAG_DHCP_SERVERID: bcopy(&cp[i+2],&cl->serv_addr.s_addr,4); break;
		case TAG_DHCP_IP: bcopy(&cp[i+2],&cl->req_addr.s_addr,4); break;
		case TAG_MAX_MSGSZ: 
			/* Note that as per spec, this is a 2-byte unsigned quantity.  We are storing it in an 'unsigned int'. */
			{	unsigned short temp_cl_max_len;
				bcopy(&cp[i+2],&temp_cl_max_len,2); 
				cl->max_len = (unsigned int) ntohs(temp_cl_max_len);
			}
			break;
		case TAG_DHCP_CLIID:
			cl->cliid = (struct shared_bindata*) smalloc(sizeof(struct shared_bindata)+cp[i+1]-1);
			bzero(cl->cliid,sizeof(struct shared_bindata)+cp[i+1]-1);
			bcopy(&cp[i+2],cl->cliid->data,cp[i+1]);
			cl->cliid->length=cp[i+1];
			cl->cliid->linkcount=1;
			cl->cliid_explicit=TRUE;	/* we didn't have to infer it */
			/* rule out some common bogus Client IDs for htype=10Mb-Ethernet */
			if (cl->cliid->length  == haddrlength(HTYPE_ETHERNET)+1) {
				if (!bcmp((char *) &cl->cliid->data, (char *) &htype_ethernet, 1)) {
					/* rule common bogus Client ID addresses of 0x01000000000000 */
					if (!bcmp((char *) &cl->cliid->data[1], (char *) &enet_all_zeros, haddrlength(HTYPE_ETHERNET))) {
						if (debug > 2)
							report(LOG_NOTICE, "  bad Client ID option (all zeros); ignoring DHCP client (chaddr=%s)", haddrtoa(pkt->bp_chaddr, pkt->bp_hlen));
						return 1;
					}
					/* rule out bogus Client ID corresponding to broadcast or multicast Ethernet addresses */
					if ( cl->cliid->data[1] & 0x01 ) {
						if (debug > 2)
							report(LOG_NOTICE, "  bad Client ID option (corresponds to broadcast or multicast Ethernet address); ignoring DHCP client (chaddr=%s)", haddrtoa(pkt->bp_chaddr, pkt->bp_hlen));
						return 1;
					}
				}
			}
			break;
		case TAG_DHCP_IPLEASE:
			bcopy(&cp[i+2],&cl->secs, 4);
			cl->secs = ntohl(cl->secs);
			break;
		case TAG_DHCP_REQ:
			cl->req_opts=(unsigned char*)smalloc(cp[i+1]+1);
			bcopy(&cp[i+1],cl->req_opts,cp[i+1]+1);
			break;
		default:break;
		}
		i++;
		i+=cp[i]+1;
	}

	if (cl->cliid) {
		report(LOG_ERR, "  error parsing DHCP packet, ignoring DHCP client (cliid=%s, but may be inaccurate)", cliid_btoa(cl->cliid));
	} else if (cl->cliidhw) {
		report(LOG_ERR, "  error parsing DHCP packet, ignoring DHCP client (chaddr=%s,  but may be inaccurate)", cliid_btoa(cl->cliidhw));
	} else 
		report(LOG_ERR, "  error parsing DHCP packet, ignoring DHCP client (no cliid nor chaddr available)");
	return 1; 
}


boolean ipcmp(a, b)
hash_datum *a,*b;
{
char *key;
struct host *data;
	/* something's probably very wrong is a or b are NULL */
	if ((a==NULL) || (b==NULL))
		return FALSE;
	key=(char *)a;
	data=(struct host *)b;
	return (((struct in_addr*)key)->s_addr==data->iaddr.s_addr);
}
boolean cliidhwlookcmp(a, b)
hash_datum *a,*b;
{
struct shared_bindata *key;
struct host *data;
	key=(struct shared_bindata *)a;
	data=(struct host *)b;
	if(data->htype!=key->data[0])
		return FALSE;

	if(bcmp(data->haddr,&key->data[1],key->length-1))
		return FALSE;
	return TRUE;
}
boolean cliidcmp(a,b)
hash_datum *a,*b;
{
struct shared_bindata *key;
struct host *data;
	key=(struct shared_bindata *)a;
	data=(struct host *)b;
	if(data->flags.cliid)
		return (key->length==data->cliid->length && 
			!bcmp(key->data,data->cliid->data,key->length));
	else
		return FALSE;
}
boolean boundcliidcmp(a,b)
hash_datum *a,*b;
{
struct shared_bindata *key;
struct host *data;
	key=(struct shared_bindata *)a;
	data=(struct host *)b;
	return (key->length==data->cl->length && 
			!bcmp(key->data,data->cl->data,key->length));
}
	

/* recover the memory allocated to struct client_info * cl */
void free_client_info(cl)
struct client_info *cl;
{

#ifdef DEBUG
	if (debug > 15) {
		report(LOG_DEBUG, "free_client_info() called");
		report_cliid(LOG_DEBUG, cl->cliid);
	}
#endif

if (!cl) {
	/* this *really* shouldn't ever happen */
	report(LOG_ERR,"free_client_info(): called with NULL pointer! (will not free memory)");
	return;
}

   /* We must call del_bindata() on the structures pointed to by cliid, cliidhw,
      vend, and user fields, because each of those has its OWN linkcount to keep
      track of the number sharers of that data, which del_bindata() updates/frees.
	  We HOPE everyone using any of these fields correctly increments the linkcounter,
	  otherwise we'll free the structure prematurely.  We HOPE everyone no longer using
	  any of these fields calls del_bindata(), or we'll leak memory.  (I suspect the
	  latter is happening, haven't seen the former happen yet.)
	*/
	del_bindata(cl->cliid);
	del_bindata(cl->cliidhw);
	del_bindata(cl->vend);
	del_bindata(cl->user);

	/* these fields are not shared */
	if (cl->msg) 
		free(cl->msg);
	if (cl->req_opts)
		free(cl->req_opts);

}


void dhcp_discover(pkt)
struct bootp *pkt;
{
	struct client_info cl;
	struct host *hp, *hp2, *hp3, *registered;
	unsigned int32 dest; /* network byte order */
	int len;
	unsigned iphashcode,hwhashcode,cablehashcode,rc;
	struct opts opt;
	struct lastbind *oldbind;
	int32 bootsize=0;
	int exists;		/* is the binding we choose to offer is one that already exists for this client? */
	struct cable *c;
	boolean inconsistent;
	int roaming=0;
	struct flag sent_options;

	bzero(&cl,sizeof(cl));
	cl.max_len=576; 	/* for testing overload functions, temporarily reduce this, e.g. to 300 */

	bzero(&sent_options,sizeof(sent_options));

	if(debug>1)
		report(LOG_INFO," processing DHCPDISCOVER");
	discovers_received++;
	rc=dhcp_opts(pkt,&cl);

	/* we don't need msg here, dhcp_opts() already produces one
	if(!cl.cliid)
		report(LOG_ERR,"  couldn't determine client identifier, ignoring dhcpdiscover");
	*/
	if(rc){
		free_client_info(&cl);
		return;
	}

	if(cl.msg)free(cl.msg);

	if(debug>1)
		report_cliid(LOG_INFO,cl.cliid);

	cl.clhashcode=hash_HashFunction(cl.cliid->data,cl.cliid->length);

	/* We look for an IP address to offer, checking databases in the following order.  In each case,
	   if a potential IP address is found, we check if it is appropriate; if it is not, we discard it and
	   continue our search.  
	   The search order is:
			static IP address assigned in bootptab for this client's clientid
			current binding for this client's clientid
			previous binding for this client's clientid
			requested IP address option in DISCOVER packet, if present
			a new dynamic IP address

		Once we find an IP address to offer, we actually bind the client to this IP address, with
			a very short expiration time.
		If we find no IP address to offer, we are silent.

		Note that we violate spec, which says you search for current or oldbinding first.
		We violate because spec would prevent us from assigning a static IP address to
		a client that previous was bound to a dynamic address, when both are now on the same net.
	*/


	registered=NULL;

	/* Look for clientid in bootptab.  (Remember that if incoming packet doesn't specify a 
	   DHCP Client ID option, we've computed its ClientID by concatenating its htype with its haddr.) */
	hp =(struct host *)hash_Lookup(clhashtable, cl.clhashcode,cliidcmp,cl.cliid);

#ifdef RETRY_HADDR_AS_CLIID
	/* I've ifdef'd out this because pretending to find a cliid match for clients when there's a mismatch 
	   between the packet's Cliid and the bootptab's cliid can cause problems later as we start storing
	   the inconsistent values in our hashtables.  Besides, we probably shouldn't be pretending to find
	   a match in the first place -- errors in the cliid specified by the bootptab or the client
	   should be corrected, not ignored.
	*/
	if(!hp){ 
		/* We didn't find ClientID in bootptab's clhashtable using ClientID as key.  
		   (Recall that clhashtable contains bootptab entries keyed by cliid (if 'cl' explicitly 
		   specified in bootptab entry) else bootptab entries keyed by haddr.) 
		   Perhaps the incoming packet contains a DHCP ClientID option with a garbage value.
		   Or perhaps the bootptab entry specifies a garbage cliid (via an incorrect 'cl' tag).
		   In either case, we may have more luck if we look up the incoming packet's haddr in
		   bootptab's hwhastable using haddr as key.  
		*/
		hwhashcode=hash_HashFunction(cl.cliidhw->data+1,cl.cliidhw->length-1);
		hp =(struct host *)hash_Lookup(hwhashtable, hwhashcode, cliidhwlookcmp,cl.cliidhw);
	}
#endif

	if (hp) { /* found in bootptab */
		registered=hp;
		roaming = hp->flags.roaming;
		if (hp->iaddr.s_addr!=0) {
			if (compact_messages && debug) {
				char giaddrbuf[MAX_INETADDR_STRING_LEN];
				char iaddrbuf[MAX_INETADDR_STRING_LEN];

				strncpy(giaddrbuf, inet_ntoa(pkt->bp_giaddr), MAX_INETADDR_STRING_LEN-1);
				strncpy(iaddrbuf, inet_ntoa(hp->iaddr), MAX_INETADDR_STRING_LEN-1);
				report(LOG_INFO, "DHCPDISCOVER %s [%s]: found %s (%s)", 
					haddrtoa2(pkt->bp_chaddr, pkt->bp_hlen), giaddrbuf, iaddrbuf, hp->hostname->string);
			}
			if (debug > 3)
				report(LOG_INFO, "  client id/haddr found in bootptab with static IP address %s", inet_ntoa(hp->iaddr));
			if (check_net(&(pkt->bp_giaddr),&(hp->iaddr)))  {
				if (debug > 3) 
					report(LOG_INFO,"  static IP address is OK for client's network");
			}
			else {
				if (debug > 3) 
					report(LOG_INFO,"  static IP Address not appropriate for client's network, will not use");
				hp=NULL;
			}
		} else {
			if (debug > 3)
				report(LOG_INFO, "  client id/haddr found in bootptab with no static IP address, will not use this bootptab entry");
			hp=NULL;
		}
	} else {
		if (compact_messages && debug)
			report(LOG_INFO, "DHCPDISCOVER %s [%s]: not found", 
				haddrtoa2(pkt->bp_chaddr, pkt->bp_hlen), inet_ntoa(pkt->bp_giaddr));
		if (debug > 3) 
			report(LOG_INFO, "  client id/haddr not found in bootptab");
	}


	/* check if there is a current binding for this cliid (and possibly ip address, if static found in bootptab) */
	exists=FALSE;	/* assume client has no current binding */
	hp2 =(struct host *)hash_Lookup(boundhshtbl,cl.clhashcode,boundcliidcmp,cl.cliid);

	if (hp) { /* we had found an appropriate static IP address in bootptab */
		if (!hp2)  { 
			if (debug > 3)
				report(LOG_INFO, "  client id not presently bound");
		} else { 
			if (debug > 3)
				report(LOG_INFO, "  client id presently bound to %s", inet_ntoa(hp2->iaddr));
			if (!hp2->flags.expire) {
				if (debug > 3 || log_expirations)
					report(LOG_INFO, "  current binding was assigned via BootP (not DHCP), freeing binding");
				reuse_addr(hp2);
				hp2=NULL;
			} else { /* current binding was assigned via DHCP, not BootP */
				if (hp2->expire <= now) {
					if (debug > 3 || log_expirations)
						report(LOG_INFO, "  current binding already expired (awaiting garbage collection), freeing binding");
					reuse_addr(hp2);
					hp2=NULL;
				} else { /* current binding has not yet expired */
					if (hp->iaddr.s_addr == hp2->iaddr.s_addr) {
					exists=TRUE; /* we passed the entire chain of tests leading to this! */
						if (debug > 3) {
							report(LOG_INFO, "  binding has same IP address as bootptab entry; OK to keep using binding");
						}
					} else {
						if (debug > 3 || log_expirations)
							report(LOG_INFO, "  binding has different IP address as bootptab entry, freeing binding");
						reuse_addr(hp2);
						hp2 = NULL;
					}
				}
			}
		}
		/* make sure ipboundhshtbl agrees with state of hp2 */
		iphashcode=hash_HashFunction((u_char*)&hp->iaddr.s_addr,4);
		hp3 =(struct host *)hash_Lookup(ipboundhshtbl,iphashcode,ipcmp,(hash_datum *)&(hp->iaddr.s_addr));
		if (hp2) { /* cliid has current binding */
			if (hp3) { /* ip address statically-assigned to cliid is currently bound */
				/* make sure existing entry in hp3 agrees with hp2 */
				if (!BINDATA_EQ(hp3->cl, cl.cliid)) {
					report(LOG_ERR, "dhcp_discover(): the following cliid...");
					report_cliid(LOG_ERR, hp->cl);
					report(LOG_ERR, "was assigned a static IP (%s) in bootptab, cliid was found bound to this IP in boundhshtbl...", inet_ntoa(hp->iaddr));
					report(LOG_ERR, "but this IP found in ipboundhshtbl bound to a *DIFFERENT* cliid (cliid=%s)", cliid_btoa(hp3->cl));
					report(LOG_ERR, "freeing both bindings");
					reuse_addr(hp3);
					reuse_addr(hp2);
					hp3 = hp2 = NULL;
				} else
					exists=TRUE;
			} else { /* ip address statically-assigned to cliid is not currently bound */
				report(LOG_ERR, "dhcp_discover(): cliid (%s) assigned a static IP (%s) in bootptab, cliid found bound to this IP in boundhshtbl, but this IP *NOT* found in ipboundhshtbl!", cliid_btoa(hp->cl), inet_ntoa(hp->iaddr));
				report(LOG_ERR, "freeing binding found in boundhshtbl");
				reuse_addr(hp2);
				hp2=NULL;
			}
		} else { /* cliid does not have current binding */
			if (hp3) { /* ip address statically-assigned to cliid is currently bound */
				report(LOG_ERR, "dhcp_discover(): cliid (%s) assigned a static IP (%s) in bootptab, cliid not found in boundhshtbl, but this IP *FOUND* in ipboundhshtbl!", cliid_btoa(hp->cl), inet_ntoa(hp->iaddr));
				report(LOG_ERR, "the unexpected entry found in ipboundhshtbl had the cliid %s", cliid_btoa(hp3->cl));
				report(LOG_ERR, "freeing unexpected binding found in ipboundhshtbl");
				reuse_addr(hp3);
				hp3=NULL;
			} else { /* ip address statically-assigned to cliid is not currently bound */
				/* this is normal */
			}
		}
	} 

	

	if (!hp) { /* we didn't find an appropriate static IP address in bootptab */
		if (hp2) { /* client is presently bound */
			hp=hp2;
			hp2=NULL;
			exists = TRUE;
			/* Check that current binding is still OK to re-use */
			if (debug > 3)
				report(LOG_INFO,"  client id is currently bound to %s, checking if ok to re-use", inet_ntoa(hp->iaddr));
			/* check that hp->iaddr is still in bootptab */
			iphashcode=hash_HashFunction((u_char*)&hp->iaddr.s_addr,4);
			hp2=(struct host *)hash_Lookup(iphashtable,iphashcode,ipcmp, (hash_datum*)&(hp->iaddr.s_addr));
			if (!hp2) {
				if (debug > 3 || log_expirations) {
					report(LOG_INFO,"  IP address no longer in bootptab, will not re-use");
					report(LOG_INFO,"  freeing binding");
				}
				reuse_addr(hp);
				hp = NULL;
				exists = FALSE;
			}
			if (hp) {
				/* check that the binding is also in ipboundhshtbl - it better be! */
				hp3 =(struct host *)hash_Lookup(ipboundhshtbl,iphashcode,ipcmp,(hash_datum *)&(hp->iaddr.s_addr));
				if (!hp3) { /* this should not happen! */
					report(LOG_ERR, "dhcp_discover(): cliid (%s) found in boundhshtbl, but corresponding IP (%s) *NOT* found in ipboundhshtbl!", cliid_btoa(hp->cl), inet_ntoa(hp->iaddr));
					report(LOG_ERR, "freeing binding");
					reuse_addr(hp);
					hp2 = hp = NULL;
					exists=FALSE;
				}
			}
				if (hp && hp3 && !hp3->flags.cliid) {
				report(LOG_ERR, "dhcp_discover(): cliid (%s) found in boundhshtbl, corresponding IP (%s) found in ipboundhshtbl, but latter lacks a cliid!", cliid_btoa(hp->cl), inet_ntoa(hp3->iaddr));
				report(LOG_ERR, "freeing both bindings");
				reuse_addr(hp3);
				reuse_addr(hp);
				hp3 = hp = NULL;
				exists=FALSE;
			}
			if (hp && hp3 && !BINDATA_EQ(cl.cliid, hp3->cliid)) {
				report(LOG_ERR, "dhcp_discover(): cliid (%s) found in boundhshtbl, corresponding IP (%s) found in ipboundhshtbl,...", cliid_btoa(hp->cl), inet_ntoa(hp3->iaddr));
				report(LOG_ERR, "but latter's cliid (%s) does not match!", cliid_btoa(hp3->cl));
				report(LOG_ERR, "freeing both binding");
				reuse_addr(hp3);
				reuse_addr(hp);
				hp3 = hp = NULL;
				exists=FALSE;
			}
			if (hp && !hp->flags.expire) {
				if (debug > 3 || log_expirations) {
					report(LOG_INFO, "  current binding was assigned via BootP (not DHCP), will not re-use");
					report(LOG_INFO, "  freeing binding");
				}
				reuse_addr(hp);
				hp=NULL;
				exists = FALSE;
			}
			if (hp && (hp->expire <= now)) {
				if (debug > 3 || log_expirations) {
					report(LOG_INFO, "  current binding already expired (awaiting garbage collection)");
					report(LOG_INFO, "  freeing binding");
				}
				reuse_addr(hp);
				hp=NULL;
				exists = FALSE;
			}
			if (hp) {
				/* If bootptab entry has a cliid, make sure it matches the requester's */
				if (hp2 && hp2->flags.cliid && !BINDATA_EQ(hp2->cliid,cl.cliid)) {
						report(LOG_ERR,"  bootptab entry for this IP has different cliid! Something's very wrong!");
						report(LOG_ERR,"  freeing binding");
						reuse_addr(hp);
						hp=NULL;
						exists = FALSE;
				}
			}
			if(hp) { 
				/* Make sure IP address is still appropriate, given client's present wire. */
				if (!check_net(&(pkt->bp_giaddr),&(hp->iaddr))) {
					if (debug > 3 || log_expirations) {
						report(LOG_INFO,"  IP address no longer appropriate (client on different network); will not re-use");
						report(LOG_INFO,"  freeing binding");
					}
					reuse_addr(hp);
					hp=NULL;
					exists = FALSE;
				}
			}
			if (hp && hp2->flags.dynamic) { /* If bootptab entry is a dynamic IP address */
				/* Make sure the dynamic address is a member of a dynamic pool we're handling */
				if (!hp->flags.cable) {
					report(LOG_ERR, "  IP %s is bound, and bootptab indicates it is dynamic, but it is *not* a member of any pool!  Something's very wrong!", inet_ntoa(hp->iaddr));
					report(LOG_ERR, "  freeing binding"); 
					reuse_addr(hp);
					hp=NULL;
					exists = FALSE;
				}
				/* Make sure the IP address is still appropriate, given the client's present wire, and
					policy restrictions. */
				/* Note we cannot use use_addr() since it will return FALSE when proposed IP address
				   is not free, and we *know* this address is not free, since it is bound to us.  We'll have
				   to go down a level and use access_network(), relying on the fact that we've already
				   taken care of doing check_net().  */
				cablehashcode = hash_HashFunction((u_char*)hp2->cable->string, strlen(hp2->cable->string));
    			c = (struct cable *) hash_Lookup(cablehshtbl, cablehashcode, cablelcmp, hp2->cable->string);
    			if (c) { /* only test if we have an access policy defined for this cable */
					if (roaming)
						rc=access_network(c->code,ROAMING_ACCESS|REG_ACCESS|DHCP_ACCESS);
					else if (registered)
						rc=access_network(c->code,REG_ACCESS|DHCP_ACCESS);
					else
						rc=access_network(c->code,DHCP_ACCESS);
					if (!rc) {
						if (debug > 3 || log_expirations) {
							report(LOG_INFO,"  dynamic IP Address disallowed by access policy; will not re-use");
							report(LOG_INFO,"  freeing binding");
						}
						reuse_addr(hp);
						hp=NULL;
						exists = FALSE;
					}
				}
			}
		} else if (debug > 3) report(LOG_INFO, "  client not currently bound");
	}


	if(!hp){
		oldbind=(struct lastbind*)hash_Lookup(lastbindhshtbl,cl.clhashcode,lastbindcmp,&cl);
		if(oldbind){
			if(debug>3)
				report(LOG_INFO, "  Found previous binding to %s, checking if ok to re-use",inet_ntoa(oldbind->last_ip));
			/* check that hp->iaddr is still in bootptab */
			iphashcode=hash_HashFunction((u_char*)&oldbind->last_ip.s_addr,4);
			hp=(struct host *)hash_Lookup(iphashtable,iphashcode,ipcmp, (hash_datum*)&(oldbind->last_ip.s_addr));
			if (!hp) {
				if (debug > 3)
					report(LOG_INFO, "  IP Address no longer in bootptab, will not re-use");
				hp = NULL;
			}
			/* Make sure this IP address not presently bound.  If it is, we know it is to someone
			   else, since by now we now this client is not presently bound. */
			if (hp && hash_Lookup(ipboundhshtbl, hp->iphashcode, ipcmp, (hash_datum *)&(hp->iaddr.s_addr))) {
				if (debug >3)
					report(LOG_INFO, "  IP address now bound to someone else, will not re-use");
					hp = NULL;
			}
			/* If bootptab entry has a cliid, make sure it matches the requester's */
			if (hp && hp->flags.cliid && !BINDATA_EQ(hp->cliid,cl.cliid)) {
				if (debug > 3)
					report(LOG_INFO, "  Bootptab entry for this IP has different cliid, will not re-use");
				hp=NULL;
			}
			if (hp && hp->flags.dynamic) { /* If bootptab entry is a dynamic IP address */
				/* Make sure the IP address is still appropriate, given the client's present wire, and
					policy restrictions. Note that use_addr(), if sucessful, will actually allocate
				    the address to us -- that is, remove it from pool of available dynamic addrs. */
				if (roaming)
					hp=use_addr(hp,pkt->bp_giaddr,ROAMING_ACCESS|REG_ACCESS|DHCP_ACCESS,0);
				else if (registered)
					hp=use_addr(hp,pkt->bp_giaddr,REG_ACCESS|DHCP_ACCESS,0);
				else
					hp=use_addr(hp2,pkt->bp_giaddr,DHCP_ACCESS,0);
				if (!hp) {
					if (debug > 3)
						report(LOG_INFO, "  Dynamic IP Address no longer appropriate (client on different network, or access policy change); will not re-use");
					hp=NULL;
				}
			}
			if(hp && ! hp->flags.dynamic) { /* If bootptab is a static IP address */
				/* Make sure IP address is still appropriate, given client's present wire. */
				if (!check_net(&(pkt->bp_giaddr),&(hp->iaddr))) {
					if (debug > 3)
						report(LOG_INFO, "  Static IP Address no longer appropriate (client on different network); will not re-use)");
					hp=NULL;
				}
			}
		} else if (debug > 3) report(LOG_INFO, "  No previous binding found");
	}

	if(!hp){
		if(cl.req_addr.s_addr==0){
			if (debug > 3)
				report(LOG_INFO, "  Client did not specify Requested IP address option");
		} else {
			if (debug > 3)
				report(LOG_INFO, "  Client specified option: Requested IP address %s", inet_ntoa(cl.req_addr));
			iphashcode=hash_HashFunction((u_char*)&cl.req_addr.s_addr,4);
			hp=(struct host *)hash_Lookup(iphashtable,iphashcode,ipcmp, (hash_datum*)&(cl.req_addr.s_addr));
			if (!hp) {
				if (debug > 3)
					report(LOG_INFO,"  Requested IP Address not in bootptab, will not use");
				hp = NULL;
			}
			/* Make sure this IP address not presently bound.  If it is, we know it is to someone
			   else, since by now we now this client is not presently bound. */
			if (hp && hash_Lookup(ipboundhshtbl, hp->iphashcode, ipcmp, (hash_datum *)&(hp->iaddr.s_addr))) {
				if (debug >3)
					report(LOG_INFO, "  IP address already bound to someone else, will not re-use");
					hp = NULL;
			}
			/* If bootptab entry has a cliid, make sure it matches the requester's */
			if (hp && hp->flags.cliid && !BINDATA_EQ(hp->cliid,cl.cliid)) {
				if (debug > 3) 
					report(LOG_INFO, "  Bootptab entry for requested IP address has different cliid, will not use");
				hp=NULL;
			}
			if (hp && hp->flags.dynamic) { /* If bootptab entry is a dynamic IP address */
				/* Make sure the IP address is still appropriate, given the client's present wire, and
					policy restrictions. Note if use_addr() succeds, it allocates the address to us,
				    removing it from pool of available dynamic addrs. */
				if (roaming) 
					hp=use_addr(hp,pkt->bp_giaddr,ROAMING_ACCESS|REG_ACCESS|DHCP_ACCESS,0);
				else if(registered)
					hp=use_addr(hp,pkt->bp_giaddr,REG_ACCESS|DHCP_ACCESS,0);
				else
					hp=use_addr(hp2,pkt->bp_giaddr,DHCP_ACCESS,0);
				if (!hp) {
					if (debug > 3)
						report(LOG_INFO, "  Requested dynamic IP Address not appropriate (client on different network, or access policy restriction); will not use");
					hp=NULL;
				}
			}
			if(hp && ! hp->flags.dynamic) { /* If bootptab is a static IP address */
				/* Make sure IP address is still appropriate, given client's present wire. */
				if (!check_net(&(pkt->bp_giaddr),&(hp->iaddr))) {
					if (debug > 3)
						report(LOG_INFO, "  Requested static IP Address not appropriate (client on different network); will not use");
					hp=NULL;
				}
			}
		}
	}


	/* If nothing found yet, try to assign new dynamic IP address  */
	if(!hp){  /* first try based on whether client is registered in bootptab */
		if (roaming) {
			if (debug > 6)
				report(LOG_INFO, "  Searching for a dynamic IP address for this roaming client...");
			hp=get_dynamic_ip(pkt->bp_giaddr,ROAMING_ACCESS|REG_ACCESS|DHCP_ACCESS);
		} else if(registered) {
			if (debug > 6)
				report(LOG_INFO, "  Searching for a dynamic IP address for this registered client...");
			hp=get_dynamic_ip(pkt->bp_giaddr,REG_ACCESS|DHCP_ACCESS);
		} else {
			if (debug > 6)
				report(LOG_INFO, "  Searching for a dynamic IP address for this unregistered client...");
			hp=get_dynamic_ip(pkt->bp_giaddr,DHCP_ACCESS);
		}
	}

	if(!hp){
		if (compact_messages && debug)
			report(LOG_INFO, "DHCPDISCOVER %s [%s]: no offer", haddrtoa2(pkt->bp_chaddr, pkt->bp_hlen), inet_ntoa(pkt->bp_giaddr));
		if (debug > 3)
			report(LOG_INFO,"  No appropriate dynamic ip addresses available, ignoring client");
		free_client_info(&cl);
		return;
	}

	/* At this point, we think we have a good ip address to offer in 'hp'. */

	if (compact_messages && debug) {
		char giaddrbuf[MAX_INETADDR_STRING_LEN];
		char iaddrbuf[MAX_INETADDR_STRING_LEN];

		strncpy(giaddrbuf, inet_ntoa(pkt->bp_giaddr), MAX_INETADDR_STRING_LEN-1);
		strncpy(iaddrbuf, inet_ntoa(hp->iaddr), MAX_INETADDR_STRING_LEN-1);
		report(LOG_INFO, "DHCPDISCOVER %s [%s]: offer %s (%s)",
			haddrtoa2(pkt->bp_chaddr, pkt->bp_hlen), giaddrbuf, iaddrbuf, hp->hostname->string);
	}
	if (debug > 1)
		report(LOG_INFO, "  Selected IP address %s (%s) to offer",inet_ntoa(hp->iaddr),
			hp->hostname->string);

	if (!exists) { /* client not presently bound, create a new struct host to contain the binding, using the
					   'hp' that we've located as a template.  (If client already bound, we'll keep using the
					   struct host that is presently bound, updating it as necessary.) */
		if (!(hp=create_stub(hp,&cl))) {	/* create the new struct, fill in a few values */
			report(LOG_ERR, "   dhcp_discover(): create_stub() returned NULL, cannot create new binding");
			/* reuse_addr(hp); we can't call this since hp is toast.  We leak the IP address if dynamic, but no memory */
			free_client_info(&cl);
			return;
		}
		if (!create_lease(hp, cl.cliid, cl.vend, cl.user, 0)) {  /* fills in most values in hp */
			report(LOG_ERR, "   dhcp_discover(): create_lease() returned NULL, cannot create new binding");
			hp->linkcount = 1; /* to force free_host() to free it, as it starts =0 */
			reuse_addr(hp);
			free_client_info(&cl);
			return;
		}
		if(hp->flags.dynamic)
			hp->code=DHCP_FLAG;
		else
			hp->code=RESERVED_FLAG;
		hp->xid=pkt->bp_xid;
		hp->linkcount+=2; /* starts =0; will insert into two hashtabs next */
		/* hp->expire and hp->flags.expire will be set below */
#ifdef DEBUG
		if (debug > 15)
			report(LOG_DEBUG, "  will insert into boundhshtbl...");
#endif
		if(!hash_Insert(boundhshtbl,cl.clhashcode, boundcliidcmp, cl.cliid, (hash_datum*)hp)) {
			leases++;
#ifdef DEBUG
			if (debug > 15)
				report(LOG_DEBUG, "  incremented leases counter to %d", leases);
#endif
        } else {
			report(LOG_ERR, "    dhcp_discover(): hash_Insert(boundhshtbl) failed (cliid=%s)", cliid_btoa(cl.cliid));
			report(LOG_ERR, "    dhcp_discover(): not increasing 'leases' counter...so count may no longer be right");
        }

		/* You can positively identify DHCP leases in the current bindings; they all have hp->flags.expire==1.
		   You can then subdivide them into statics (hp->code==RESERVED_FLAG), or dynamics (hp->code==DHCP_FLAG) */
#ifdef DEBUG
		if (debug > 15)
			report(LOG_DEBUG, "  will insert into ipboundhshtbl...");
#endif
		if (hash_Insert(ipboundhshtbl,hp->iphashcode,ipcmp,(hash_datum *)&(hp->iaddr.s_addr),(hash_datum*)hp)) {
			report(LOG_ERR, "    dhcp_discover(): hash_Insert(ipboundhshtbl) failed for IP address %s (cliid=%s)", inet_ntoa(hp->iaddr), cliid_btoa(cl.cliid));
		}

	}
	/* We set hp->expire *here*, and not inside the if(!exists) block above, so that if we get multiple
	   DHCPDISCOVERs from this client while we have an outstanding OFFER to it, the server's notion of
	   when the OFFER expires (hp->expire) continues to be updated, even though the OFFER already exists.
	   There is a side effect -- if the lease we are offering is actually one the client already had with
	   us (i.e. not just an OFFER, but a real lease), we are overwriting the old lease's hp->expire value,
	   shortening it presumably.  That's OK here, as since the client is in INIT, he clearly has discarded his
	   old lease. */
	hp->expire=now+TIME_OFFSET;	/* offer will imply a very short lease */
	hp->flags.expire=1;

	/* Report if client left some MBZ flag bits set, and fix it in reply */
	if (ntohs(pkt->bp_flags) & BP_FLAGS_MBZ) {
		if (debug > 9) {
			report(LOG_NOTICE, " some MBZ bits in flags field were set in DHCPDISCOVER (cliid=%s)", cliid_btoa(cl.cliid));
		}
		pkt->bp_flags = htons(ntohs(pkt->bp_flags) & ~ BP_FLAGS_MBZ);  /* clear MBZ bits for reply */
	}

	/* construct OFFER packet */
	rpkt=(struct bootp *)send_buf;
	bcopy(pkt,rpkt,pktlen);
	rpkt->bp_op=BOOTREPLY;
	rpkt->bp_hops=0;
	rpkt->bp_secs=0;
	rpkt->bp_yiaddr.s_addr=hp->iaddr.s_addr;
   	rpkt->bp_ciaddr.s_addr=0;
	if(fill_header(rpkt,hp,&bootsize)){
		free_client_info(&cl);
		reuse_addr(hp);
		report(LOG_ERR,"  fill_header() returned non-zero, huh?");
		return;
	}
    /* always stick our name in. */
	/* note strncpy will right-pad with nul, clearing out any garbage - good */
	strncpy(rpkt->bp_sname, hostname, BP_SNAME_LEN);	/* may be overwritten if options field overflows */

	dest=hp->flags.reply_addr?hp->reply_addr.s_addr:0;

	opt.cur=0;
	opt.len=cl.max_len;
	insert_byte_opt(rpkt,&opt,TAG_DHCP_MSG,DHCPOFFER);
	insert_ulong_opt(rpkt,&opt,TAG_DHCP_SERVERID,my_ip_addr.s_addr);
	if(hp->flags.lease) {
		insert_ulong_opt(rpkt,&opt,TAG_DHCP_IPLEASE,htonl(hp->lease));
		if (debug > 9)
			report(LOG_INFO,"  Offered lease would expire in %d seconds (%s)", hp->lease, cctime(now+hp->lease));
	} else { 
		insert_ulong_opt(rpkt,&opt,TAG_DHCP_IPLEASE,htonl(DEFAULT_LEASE));
		if (debug > 9)
			report(LOG_INFO,"  Offered lease would expire in %d seconds (%s)", DEFAULT_LEASE, cctime(now+DEFAULT_LEASE));
	}
	if(hp->flags.renew) {
		insert_ulong_opt(rpkt,&opt,TAG_DHCP_RENEW,htonl(hp->renew));
		if (debug > 9)
			report(LOG_INFO,"  Offered lease would set renewal in %d seconds (%s)", hp->renew, cctime(now+hp->renew));
	}
	if(hp->flags.rebind) {
		insert_ulong_opt(rpkt,&opt,TAG_DHCP_REBIND,htonl(hp->rebind));
		if (debug > 9)
			report(LOG_INFO,"  Offered lease would set rebind in %d seconds (%s)", hp->rebind, cctime(now+hp->rebind));
	}
	if(cl.user){
		insert_string_opt(rpkt,&opt,TAG_DHCP_USERCLASS,
				cl.user->data,cl.user->length);
	}

	/* fill in options requested by client's Parameter Request List option, if any */
    if (rc=do_req_opts(hp, cl.req_opts, (u_char *) rpkt, &opt, &sent_options)) {
		report(LOG_ERR,"  do_req_opts() returned %d, will not send DHCPACK to Client ID %s", rc, cliid_btoa(hp->cl));
		free_client_info(&cl);
		reuse_addr(hp);
		return;
	}

	/* The bootptab may contain other option values for this client.  */
	if(hp->flags.not_rq) {
		/* bootptab entry has the NR tag, saying that our DHCP replies should include any values that appear
		   in the bootptab entry, even if they were "Not Requested" by the client in a Parm Request option and
		   are not values we always return */
		if (rc=do_opts(hp, (u_char *) rpkt, &opt, &sent_options)) { 
			report(LOG_ERR,"  do_opts() returned %d, will not send DHCPACK to Client ID %s", rc, cliid_btoa(hp->cl));
			free_client_info(&cl);
			reuse_addr(hp);
			return;
		}
	}

	insert_end_opt(rpkt,&opt);

	if (debug > 9)
		report(LOG_INFO,"  Offer is good for %d seconds (%s)", TIME_OFFSET, cctime(hp->expire));

	if (debug > 12)
		report(LOG_INFO,"  Sending DHCPOFFER for %s (max_len=%d, bytes left=%d)" ,inet_ntoa(hp->iaddr),opt.len,opt.len-opt.cur);
	else if (debug > 3)
		report(LOG_INFO,"  Sending DHCPOFFER for %s" ,inet_ntoa(hp->iaddr));
	offers_sent++;
	len=300;
	if(opt.cur<VEND_OFFSET)
		len=cl.max_len;
	if(opt.cur>len)
		len=opt.cur;
	send_dhcp_reply(rpkt,len,dest);
	free_client_info(&cl);
	/* note we do not update_lastbind() for offers */
}


void dhcp_request(pkt)
struct bootp *pkt;
{
	struct client_info cl;
	struct host *hp,*hp2,*hp3,*t;
	int len;
	unsigned int32 dest; /* network byte order */
	struct opts opt;
	struct flag sent_options;
	unsigned iphashcode,hwhashcode,cablehashcode,rc;
	int32 bootsize=0;
	int registered, exists, roaming=0;
	struct cable *c;
	char expiretime[MAXSTRINGLEN];
	char str1[2*MAXSTRINGLEN], str2[MAXSTRINGLEN];
	boolean assigning=FALSE;		/* the ACK is an addr assignment for reporting purposes */

	bzero(&cl,sizeof(cl));
	bzero(&sent_options,sizeof(sent_options));
	cl.max_len=576;	/* for testing overload functions, temporarily reduce this, e.g. to 300 */
	hp2=NULL;

	if(debug>1)
		report(LOG_INFO," processing DHCPREQUEST");
	requests_received++;
	rc=dhcp_opts(pkt,&cl);

	/* we don't need msg here, dhcp_opts() already produces one
	if(!cl.cliid)
		report(LOG_ERR,"  couldn't determine client identifier, ignoring, dhcprequest");
	*/
	if(rc){
		free_client_info(&cl);
		return;
	}

	if (compact_messages && debug) {
		char giaddrbuf[MAX_INETADDR_STRING_LEN];
		char iaddrbuf[MAX_INETADDR_STRING_LEN];

		strncpy(giaddrbuf, inet_ntoa(pkt->bp_giaddr), MAX_INETADDR_STRING_LEN-1);
		strncpy(iaddrbuf, inet_ntoa(cl.req_addr), MAX_INETADDR_STRING_LEN-1);
		report(LOG_INFO, "DHCPREQUEST  %s [%s]: request %s", haddrtoa2(pkt->bp_chaddr, pkt->bp_hlen), giaddrbuf, iaddrbuf);
	}

	if(cl.msg)free(cl.msg);

	if(debug>1)
		report_cliid(LOG_INFO,cl.cliid);

	cl.clhashcode=hash_HashFunction(cl.cliid->data,cl.cliid->length);

	if (cl.serv_addr.s_addr!=0) { /* client is SELECTING */
		if (debug > 3)
			report(LOG_INFO, "  client is in SELECTING state, Server Identifier option is %s", inet_ntoa(cl.serv_addr));
		if (!is_my_ip_addr(& cl.serv_addr)) {
			if (debug > 3)
				report(LOG_INFO, "  server identifier not me, ignoring");
			/* should we also see if we have this cliid bound and if so, unbind? */
			free_client_info(&cl);
			return;
		} else if (debug > 3)
			report(LOG_INFO, "  server identifier is me");

		if (cl.req_addr.s_addr==0) {
			if (debug > 2)
				report(LOG_ERR, "  dhcp client appears to be SELECTING, but omitted RequestedIPAddr Option, ignoring (cliid=%s)", cliid_btoa(cl.cliid));
			free_client_info(&cl);
			return;
		}



		/* locate client in current bindings */
		hp =(struct host *)hash_Lookup(boundhshtbl,cl.clhashcode,boundcliidcmp,cl.cliid);

		if (hp) { /* we found client in our current bindings */
			if (debug > 3)
				report(LOG_INFO, "  found current lease or offer for cliid");

			if (!hp->flags.expire) {
				if (debug > 3 || log_expirations) {
					report(LOG_INFO, "  current binding was assigned via BootP (not DHCP), will not use");
					report(LOG_INFO, "  freeing binding");
				}
				reuse_addr(hp);
				hp=NULL;
				send_dhcpnack(pkt,cl.cliid,"Lease wasn't offerred",cl.max_len,cl.cliid_explicit);
				free_client_info(&cl);
				return;
			}
			
			if (hp->expire <= now) {
				if (debug > 3 || log_expirations) {
					report(LOG_INFO, "  current binding already expired (awaiting garbage collection)");
					report(LOG_INFO, "  freeing binding");
				}
				reuse_addr(hp);
				hp=NULL;
				send_dhcpnack(pkt,cl.cliid,"Lease or offer already expired",cl.max_len,cl.cliid_explicit);
				free_client_info(&cl);
				return;
			}

			/* make sure the binding we found is for the same IP address as the RequestedIPAddr Option  */
			if (hp->iaddr.s_addr != cl.req_addr.s_addr) {
				if (debug > 3 || log_expirations) {
                    /* must build message up, since inet_ntoa shouldn't be called twice in same expr */
                    sprintf(str1, "  client's current binding in boundhshtbl is for %s, ", inet_ntoa(hp->iaddr));
                    sprintf(str2, "but client's RequestedIPAddr=%s, will not use", inet_ntoa(cl.req_addr));
                    strcat(str1,str2);
                    report(LOG_INFO, str1);
					report(LOG_INFO, "  freeing binding");
				}
				reuse_addr(hp);
				hp=NULL;
				send_dhcpnack(pkt,cl.cliid,"Lease wasn't offerred",cl.max_len,cl.cliid_explicit);
				free_client_info(&cl);
				return;
			}

			/* make sure we have this in the ipboundhshtbl as well, and it matches */
			iphashcode=hash_HashFunction((u_char*)&hp->iaddr.s_addr,4);
			hp2 =(struct host *)hash_Lookup(ipboundhshtbl,iphashcode,ipcmp,(hash_datum *)&(hp->iaddr.s_addr));
			if (!hp2) { /* this should not happen */
				report(LOG_ERR, "  dhcp_request(): SELECTING, cliid (%s) found in boundhshtbl with same ReqIPAddr (%s) , but ipboundhshtbl has no entry for this IP!", cliid_btoa(cl.cliid), inet_ntoa(hp->iaddr));
				report(LOG_ERR, "freeing binding");
				reuse_addr(hp);
				free_client_info(&cl);
				return;
			}
			if (!hp2->flags.cliid) {
				report(LOG_ERR, "  dhcp_request(): SELECTING, cliid (%s) found in boundhshtbl with same ReqIPAddr (%s), but entry in ipboundhshtbl has no cliid!", cliid_btoa(cl.cliid), inet_ntoa(hp->iaddr));
				report(LOG_ERR, "freeing both bindings");
				reuse_addr(hp2);
				reuse_addr(hp);
				free_client_info(&cl);
				return;
			}
			if (!BINDATA_EQ(hp2->cl, cl.cliid)) {
				report(LOG_ERR, "  dhcp_request(): SELECTING, cliid (%s) found in boundhshtbl with same ReqIPAddr (%s)...", cliid_btoa(cl.cliid), inet_ntoa(hp->iaddr));
				report(LOG_ERR, "but entry in ipboundhshtbl for this IP has different cliid (%s)", cliid_btoa(hp2->cl));
				report(LOG_ERR, "freeing both bindings");
				reuse_addr(hp2);
				reuse_addr(hp);
				free_client_info(&cl);
				return;
			}	

			exists=TRUE; /* was found in boundhshtbl and ipboundhshtbl */

			assigning=TRUE;		/* we want to report a client who SELECTs our OFFER */

			/* we'll not do any further checks on validity of request vs. lease */
			/* Fall through to sending ACK below */

			if (debug > 3) 
				report(LOG_INFO, "  skipping any further checks");

			/* why not do additional checks, to catch really screwed up clients that
			  pretend to be SELECTING, but aren't? */


		} else { /* client is confused; we did not OFFER */
			if (debug > 2)
				report(LOG_ERR, "  dchp_request(): no current lease or offer for SELECTING client, sending NAK (cliid=%s)", cliid_btoa(cl.cliid));
			/* don't know if we should just ignore, or NAK */
			send_dhcpnack(pkt,cl.cliid,"Lease wasn't offerred",cl.max_len,cl.cliid_explicit);
			free_client_info(&cl);
			return;
		}

	} else { /* cl.serv_addr.s_addr==0, so client is not SELECTING */

		if (cl.req_addr.s_addr!=0) { /* INIT-REBOOT */
			if (debug > 3)
				report(LOG_INFO, "  client is in INIT-REBOOT state, Requested IP Address option is %s", inet_ntoa(cl.req_addr));

			/* make sure requested ip is on appropriate subnet */
			if (!check_net(&(pkt->bp_giaddr),&(cl.req_addr))) {
				if (debug > 1) {
					/* must build message up, since inet_ntoa shouldn't be called twice in same expr */
					sprintf(str1, "  requested IP address %s ", inet_ntoa(cl.req_addr));
					sprintf(str2, "incorrect for gateway %s, sending NAK", inet_ntoa(pkt->bp_giaddr));
					strcat(str1,str2);
					report(LOG_INFO, str1);
				}
				send_dhcpnack(pkt,cl.cliid,"Requested IP address not appropriate for your subnet",cl.max_len,cl.cliid_explicit);
				free_client_info(&cl);
				return;
			}

			/* lookup requested IP in bootptab */
			iphashcode=hash_HashFunction((u_char*)&cl.req_addr.s_addr,4);
			hp=(struct host *)hash_Lookup(iphashtable,iphashcode,ipcmp,(hash_datum*)&cl.req_addr.s_addr);

			/* ignore if requested IP not in our bootptab.  We may not send NAK as client is in INIT-REBOOT,
				so REQUEST was broadcast and may be for OK by another server */
			if (!hp) {
				if (debug > 3)
					report(LOG_INFO, "  requested IP address %s not in bootptab, ignoring", inet_ntoa(cl.req_addr));
				free_client_info(&cl);
				return;
			}

			if(!hp->flags.dynamic) { /* bootptab is for a static ip address, so make sure bootptab cliid or haddr match requester's */
				if(hp->flags.cliid) { /* bootptab entry has a cliid */
					if(!CLIID_EQ(hp->cliid,cl.cliid)) { 
						if (debug > 3) 
							report(LOG_INFO, "  client identifier for static bootptab %s entry does not match, access denied", inet_ntoa(hp->iaddr));
						send_dhcpnack(pkt,cl.cliid,"Client Identifier doesn't match yours, access denied",cl.max_len,cl.cliid_explicit);
						free_client_info(&cl);
						return;
					} 
				} else if (hp->flags.haddr && hp->flags.htype) { /* bootptab entry has a haddr */
					if((hp->htype!=cl.cliid->data[0]) || ((cl.cliid->length-1)!=haddrlength(hp->htype))
						|| bcmp(hp->haddr,cl.cliid->data+1,cl.cliid->length-1)){ 
						if (debug > 3) 
							report(LOG_INFO, "  hardware address for static bootptab entry %s doesn't match, access denied", inet_ntoa(hp->iaddr));
						send_dhcpnack(pkt,cl.cliid,"Hardware Address doesn't match yours, access denied",cl.max_len,cl.cliid_explicit);
						free_client_info(&cl);
						return;
					}
				} else report(LOG_ERR,"  bootptab static entry for %s has no cliid or haddr, huh?", inet_ntoa(hp->iaddr));
			} else { /* bootptab entry is for a dynamic address */
				/* Make sure the dynamic address is a member of a dynamic pool we're handling */
				if (!hp->flags.cable) {
					if (debug > 3)
						report(LOG_INFO, "  requested IP address %s is known, but is dynamic and not a member of a defined pool, ignoring", inet_ntoa(hp->iaddr));
					/* We do *not* NAK, because we're not authoritative for this dynamic IP address.  It's true
					   that it is in our bootptab, but since it is dynamic and is not a member of any defined pool,
					   we could never assign it to anyone.  So we behave the same as if the requested IP address was
					   not in our bootptab at all: we ignore client. */
					free_client_info(&cl);
					return;
				}
				/* Make sure access policy allows this client */
				/* Note we cannot use use_addr() since it will return FALSE when proposed IP address
				   is not free, even though the reason is it is not free is it is bound to us!  We'll have
				   to go down a level and use access_network(), relying on the fact that we've already
				   taken care of doing check_net().  */
				cablehashcode = hash_HashFunction((u_char*)hp->cable->string, strlen(hp->cable->string));
    			c = (struct cable *) hash_Lookup(cablehshtbl, cablehashcode, cablelcmp, hp->cable->string);
    			if (c) { /* only test if we have an access policy defined for this cable */
					hp2 =(struct host *)hash_Lookup(clhashtable, cl.clhashcode,cliidcmp,cl.cliid);
					if (hp2) {
						if (hp2->flags.roaming)
							rc=access_network(c->code,ROAMING_ACCESS|REG_ACCESS|DHCP_ACCESS);
						else
							rc=access_network(c->code,REG_ACCESS|DHCP_ACCESS);
					} else
						rc=access_network(c->code,DHCP_ACCESS);
					if (!rc) {
						if (debug > 3)
							report(LOG_INFO, "  dynamic IP address disallowed by access policy");
						send_dhcpnack(pkt,cl.cliid,"Dynamic address disallowed by access policy",cl.max_len,cl.cliid_explicit);
						free_client_info(&cl);
						return;
					}
					/* Since we were unable to use use_addr() above, we failed to remove this IP address from the
					   pool of available dynamic addresses for this cable.  So now we have to deal with this.
					   If the address is already bound, we don't need to do anything, since when it was bound it was
					   removed from the pool of available dynamic addresses. */
					iphashcode=hash_HashFunction((u_char*)&hp->iaddr.s_addr,4);
					hp2 =(struct host *)hash_Lookup(ipboundhshtbl,iphashcode,ipcmp,(hash_datum *)&(hp->iaddr.s_addr));
					if (!hp2) {
						/* We call with force=1, so the value of gwaddr is ignored, we pass any struct in_addr there just
						   for type checking.  While the flag check is not of interest to us here (already taken care of
						   earlier when access_network called), we have to provide a flag, so we'll go with the most permissive
						   possible version.    We also set the flag arg to include both ROAMING and REG here (giving
						   the client all relevant privs) since we've already performed the policy check with
						   a call to access_network() earlier.
						   
						   We ignore the return value, since all we're after is the side-effect of
						   removing hp->iaddr from the appropriate cable's list of avail dynamic ip addresses. */
						use_addr(hp, pkt->bp_giaddr, ROAMING_ACCESS|REG_ACCESS|DHCP_ACCESS, 1);
					}
				}
				/* If bootptab also has a static IP address assigned to this client, and it is appropriate
				   for the network to which the client is presently attached, then do NOT allow the client
				   to use the dynamic address he has requested.  Although the spec says we should let the client
				   keep using the dynamic address, in reality, the fact that the bootptab was updated to reflect
				   that the client's "home" is now this network means that we should really have the client
				   use his statically-assigned IP address on that network.  Sending a NAK to the client will force
				   him from INIT-REBOOT back to INIT, at which point we'll offer him the static address.  Note that
				   this is even true if the requested dynamic address is presently bound to this client...and when
				   we see the DHCPDISCOVER from the client, we'll also free the binding.  */
				/* Look for clientid in bootptab.  (Remember that if incoming packet doesn't specify a 
				   DHCP Client ID option, we've computed its ClientID by concatenating its htype with its haddr.) */
				hp2 =(struct host *)hash_Lookup(clhashtable, cl.clhashcode,cliidcmp,cl.cliid);
#ifdef RETRY_HADDR_AS_CLIID
				/* I've ifdef'd out this because pretending to find a cliid match for clients when there's a mismatch 
				   between the packet's Cliid and the bootptab's cliid can cause problems later as we start storing
				   the inconsistent values in our hashtables.  Besides, we probably shouldn't be pretending to find
				   a match in the first place -- errors in the cliid specified by the bootptab or the client
				   should be corrected, not ignored.
				*/
				if (! hp2) { 
					/* We didn't find ClientID in bootptab's clhashtable using ClientID as key.
					   (Recall that clhashtable contains bootptab entries keyed by cliid (if 'cl' explicitly
					   specified in bootptab entry) else bootptab entries keyed by haddr.)
					   Perhaps the incoming packet contains a DHCP ClientID option with a garbage value.
					   Or perhaps the bootptab entry specifies a garbage cliid (via an incorrect 'cl' tag).
					   In either case, we may have more luck if we look up the incoming packet's haddr in
					   bootptab's hwhastable using haddr as key.
					*/
					hwhashcode=hash_HashFunction(cl.cliidhw->data+1,cl.cliidhw->length-1);
					hp2 =(struct host *)hash_Lookup(hwhashtable, hwhashcode, cliidhwlookcmp,cl.cliidhw);
				}
#endif
				if (hp2 && hp2->flags.iaddr) { /* found clientid in bootptab, and has a static ip assigned */
					/* Note that we know that the static IP addr in bootptab is not the same as the requested IP addr,
					   since we've already determined that the requested IP addr is dynamic according to bootptab. */
					/* Is the static IP addr appropriate given the client's present network? */
					if (check_net(&(pkt->bp_giaddr),&(hp2->iaddr)))  {
						/* static IP addr is ok for client's present network, so DO NOT allow client to have the dynamic addr */
						if (debug > 3)
							report(LOG_INFO, "  dynamic address disallowed since an appropriate static address (%s) is available", inet_ntoa(hp2->iaddr));
						send_dhcpnack(pkt,cl.cliid," Requested address is dynamic, but an appropriate static IP address is now available",cl.max_len,cl.cliid_explicit);
						free_client_info(&cl);
						return;
					} else { 
						/* static not ok for client's present network, so DO allow client to have the dynamic addr */
					}
				}
			}

			/* see if IP address presently bound */
			exists=FALSE;
			iphashcode=hash_HashFunction((u_char*)&hp->iaddr.s_addr,4);
			hp2 =(struct host *)hash_Lookup(ipboundhshtbl,iphashcode,ipcmp,(hash_datum *)&(hp->iaddr.s_addr));
			if (hp2) { /* IP address is presently bound */
				if (hp2->flags.cliid) { /* binding has a cliid */
					 if (!CLIID_EQ(hp2->cliid,cl.cliid)) { /* make sure binding's cliid == requester's cliid */
						if (debug > 3)
							report(LOG_INFO, "  requested IP Address already leased to another Client ID");
						send_dhcpnack(pkt,cl.cliid,"Requested IP Address leased to someone else",cl.max_len,cl.cliid_explicit);
						free_client_info(&cl);
						return;
					}
				} else {
					report(LOG_ERR, "  IP address %s presently bound, but binding has no cliid!", inet_ntoa(hp2->iaddr));
					report(LOG_ERR, "  Freeing binding.");
					reuse_addr(hp);
					send_dhcpnack(pkt,cl.cliid,"Requested IP Address leased to unknown client",cl.max_len,cl.cliid_explicit);
					free_client_info(&cl);
					return;
				}
				/* at this point we know that that the RequestedIPAddr is presently bound to this cliid */
				if (!hp2->flags.expire) {
					if (debug > 3 || log_expirations) {
						report(LOG_INFO, "  current binding was assigned via BootP (not DHCP), will not re-use");
						report(LOG_INFO, "  Freeing binding.");
					}
					reuse_addr(hp);
					send_dhcpnack(pkt,cl.cliid,"Requested IP Address was assigned via BootP (not DHCP)",cl.max_len,cl.cliid_explicit);
					free_client_info(&cl);
					return;
				}
				/* It is possible that hp->expire >= now, meaning that the current binding has already
				   expired.  (That can happen if it expired by garbage collection hasn't happend yet to remove it.
				   Let's be safe and discard that binding (I'm not sure it is safe to extend it if it happens
				   to be just an OFFER that was never accepted, instead of a real lease).  Note that we don't
				   send a NAK here....we continue processing the REQUEST, just as if the client is not presently
				   bound. */
				if (hp2->expire <= now) {
					reuse_addr(hp2);
					hp2=NULL;
					if (debug > 3 || log_expirations) {
						report(LOG_INFO, "  current binding has expired (awaiting garbage collection)");
						report(LOG_INFO, "  Freeing binding, will proceed (treating client as unbound)");
					}
				}
			}

			/* see if cliid is presently bound */
			hp3 =(struct host *)hash_Lookup(boundhshtbl,cl.clhashcode,boundcliidcmp,cl.cliid);
			if (hp3) { /* cliid found in boundhshtbl */
				if (hp2) { /* requested IP addr found in ipboundhshtbl */
					/* make sure the IP address in the bound cliid record matches the requested IP addr */
					if (hp3->iaddr.s_addr == hp->iaddr.s_addr)  {
						/* this is normal, all matches */
						exists=TRUE;
					} else {
						/* not sure what this means.  We found cliid found, and requested IP bound, but cliid was bound
					   to a different IP.  Unbind cliid  */
						report(LOG_ERR, "  dhcp_request(): INIT-REBOOT: ReqIP (%s) found in ipboundhshtbl with correct cliid...", cl.req_addr);
						report_cliid(LOG_ERR, cl.cliid);
						report(LOG_ERR, "and cliid found in boundhshtbl, but bound to *DIFFERENT* IP address (%s)!", inet_ntoa(hp3->iaddr));
						report(LOG_ERR, "freeing both bindings");
						reuse_addr(hp3);
						reuse_addr(hp2);
						send_dhcpnack(pkt,cl.cliid,"Error",cl.max_len,cl.cliid_explicit);
						free_client_info(&cl);
						return;
					}
				} else { /* requested IP addr not found in ipboundhshtbl */
					/* We found the cliid bound, but the requested IP was not bound. 
					   That implies that the cliid is not presently bound to the requested IP.  Given that the client is
					   asking for a different IP address, he's clearly given up his old binding.  So we'll unbind the cliid. */
					if (debug > 3 || log_expirations) {
						report(LOG_INFO, "  dhcp_request(): cliid presently bound to %s...",inet_ntoa(hp3->iaddr));
						report(LOG_INFO, "  dhcp_request(): but is requesting %s; will free old binding", inet_ntoa(cl.req_addr));
					}
					reuse_addr(hp3);
					free_client_info(&cl);
					/* We ignore the client, which forces him to timeout and resend the request, or fall back to INIT.  Either way,
					   he'll send another packet.  When we get that one, we won't find the old binding (we just removed it),
					   so we'll be able to handle it appropriately.  We do *not* NAK him here, since it's possible that what
					   what he's asking for is OK. */
					return;
				}
			} else { /* cliid not found in boundhshtbl */
				if (hp2) { /* requested IP addr found in ipboundhshtbl */
					/* this should never happen, since we already know that the ipboundhshtbl record specifies this cliid */
					/* not much we can do here, other than log the error */
					report(LOG_ERR, "  dhcp_request(): INIT-REBOOT: ReqIp (%s), found in ipboundhshtbl with correct cliid...", inet_ntoa(cl.req_addr));
					report_cliid(LOG_ERR, cl.cliid);
					report(LOG_ERR, "but cliid *NOT* found in boundhshtbl!");
					report(LOG_ERR, "freeing binding");
					reuse_addr(hp2);
					send_dhcpnack(pkt,cl.cliid,"Error",cl.max_len,cl.cliid_explicit);
					free_client_info(&cl);
					return;
				} else { /* requested IP addr not found in ipboundhshtbl */
					/* this is normal; neither cliid nor req ip was bound; exists still == FALSE */
				}
			}

			assigning=TRUE; /* we want to report INIT-REBOOT assignments */

			/* all looks good, fall through to sending ACK below */
			
		} /* INIT-REBOOT */

		else { /* RENEWING/REBINDING/BOUND/confused */
			if (debug > 4) {
				report(LOG_INFO, "  Neither ServerAddr option nor RequestedIPAddr option specified,");
				report(LOG_INFO, "  ...so client must be RENEWING, REBINDING, BOUND (or confused)");
			}
			if (pkt->bp_ciaddr.s_addr==0) {
				if (debug > 2)
					report(LOG_ERR, "  client is RENEWING or REBINDING, but ciaddr==0.0.0.0, ignoring (cliid=%s)", cliid_btoa(cl.cliid));
				free_client_info(&cl);
				return;
			}
			/* At this point:
					Client is RENEWING if ((giaddr==0 && pktDstIP==myIP))
					Client is REBINDING if ((giaddr!=0) || (pktDstIP==broadcast))
					else client is in unknown state
			   However, we don't have access to pktDstIP here, (we're getting datagrams from the socket,
				not raw packets) so we can't tell for sure which case applies.
			   So we will lump them all together.   The effects are:
					Some RENEWING clients that are wrong would normally get NAKed; here we'll be silent,
					   so they get to keep using an inappropriate lease until it expires.
					Clients in unknown state will be treated as RENEWING or REBINDING.
					Our debug messages won't be as detailed as they could be.
			*/
			     
			if (pkt->bp_giaddr.s_addr!=0)  {
				if (debug > 4)
					report(LOG_INFO, "  giaddr!=0, so client must be REBINDING", inet_ntoa(pkt->bp_giaddr));
			} else if (debug > 4) {
				report(LOG_INFO, "  giaddr==0, so client must be RENEWING,");
				report(LOG_INFO, "  ...or is on local network and is REBINDING (or is confused)");
			}

			/* If we have a non-zero giaddr to work with, make sure requested address is appropriate. */
			/* Note that as per spec, if client is RENEWING, we cannot perform this test
			      as giaddr will always be 0.
			   Note also that if client is REBINDING and is on a directly-attached network,
				  we cannot perform this test as giaddr will be 0. */
			if (pkt->bp_giaddr.s_addr) {
				if (!check_net(&(pkt->bp_giaddr),&(pkt->bp_ciaddr))) {
					if (debug > 2) {
                    	/* must build message up, since inet_ntoa shouldn't be called twice in same expr */
                    	sprintf(str1, "  requested IP address (ciaddr) %s ", inet_ntoa(pkt->bp_ciaddr));
                    	sprintf(str2, "incorrect for gateway %s, sending NAK", inet_ntoa(pkt->bp_giaddr));
                    	strcat(str1,str2);
                    	report(LOG_INFO, str1);
					}
					send_dhcpnack(pkt,cl.cliid,"Requested IP address not appropriate for your current subnet",cl.max_len,cl.cliid_explicit);
					free_client_info(&cl);
					return;
				}
			}

            /* make sure ciaddr is in bootptab */
        	iphashcode=hash_HashFunction((u_char*)&pkt->bp_ciaddr.s_addr,4);
	        hp=(struct host *)hash_Lookup(iphashtable,iphashcode,ipcmp,(hash_datum*)&(pkt->bp_ciaddr.s_addr));
            if (!hp) {
				/* if we knew for SURE the client was RENEWING, we'd like to NAK, but this
				   client may be REBINDING, so the best we can do is be silent.  The RENEWING
				   clients will get to keep using their invalid leases until they expire at the client end,
					unfortunately, but at least we won't break REBINDING clients. */
				if (debug > 3) {
					report(LOG_INFO, "  client's ciaddr %s is not in bootptab", inet_ntoa(pkt->bp_ciaddr));
					report(LOG_INFO, "  Want to NAK if client is RENEWING, but client may be REBINDING, so will ignore");
				}
                free_client_info(&cl);
                return;
			}

            if(!hp->flags.dynamic) { /* bootptab entry is for a static ip address, so make sure bootptab cliid or haddr match requester's */
                if(hp->flags.cliid) { /* bootptab entry has a cliid */
					if (!CLIID_EQ(hp->cliid,cl.cliid)) { 
                    	if (debug > 3) 
							report(LOG_INFO, "  client identifier for %s bootptab static entry doesn't match, access denied", inet_ntoa(hp->iaddr));
                    	send_dhcpnack(pkt,cl.cliid,"Client ID for bootptab static entry doesn't match yours, access denied",cl.max_len,cl.cliid_explicit);
                    	free_client_info(&cl);
                    	return;
					}
                } else if(hp->flags.haddr && hp->flags.htype) { /* bootptab entry has a haddr */
					if((hp->htype!=cl.cliid->data[0]) || ((cl.cliid->length-1)!=haddrlength(hp->htype))
						|| bcmp(hp->haddr,cl.cliid->data+1,cl.cliid->length-1)){
						if (debug > 3) 
							report(LOG_INFO, "  hardware address for %s bootptab static entry doesn't match, access denied", inet_ntoa(hp->iaddr));
						send_dhcpnack(pkt,cl.cliid,"Hardware Address for bootptab static entry doesn't match yours, access denied",cl.max_len,cl.cliid_explicit);
						free_client_info(&cl);
						return;
					}
				} else report(LOG_ERR, "  Bootptab static entry for %s has no cliid or haddr, huh?", inet_ntoa(hp->iaddr));

            } else { /* bootptab entry is for dynamic address */ 
				/* Make sure the dynamic address is a member of a dynamic pool we're handling */
				if (!hp->flags.cable) {
					/* We're not authoritative for this dynamic IP address .  It's true that it is in our
					   our bootptab, but since it is dynamic and is not a member of any defined pool, we could 
						never assign it to anyone.  So we behave the same as if the ciaddr was not in our bootptab at all. */
					if (debug > 3)
						report(LOG_INFO, "  client ciaddr %s is known, but is dynamic and not a member of a defined pool", inet_ntoa(hp->iaddr));
						report(LOG_INFO, "  Want to NAK if client is RENEWING, but client may be REBINDING, so will ignore");
					free_client_info(&cl);
					return;
				}
				/* make sure access policy allows this client */
				/* Note we cannot use use_addr() since it will return FALSE when proposed IP address
				   is not free, even though the reason is it is not free is it is bound to us!  We'll have
				   to go down a level and use access_network(), relying on the fact that we've already
				   taken care of doing check_net(). */
				cablehashcode = hash_HashFunction((u_char*)hp->cable->string, strlen(hp->cable->string));
    			c = (struct cable *) hash_Lookup(cablehshtbl, cablehashcode, cablelcmp, hp->cable->string);
    			if (c) { /* only test if we have an access policy defined for this cable */
					hp2 =(struct host *)hash_Lookup(clhashtable, cl.clhashcode,cliidcmp,cl.cliid);
					if (hp2) {
						if (hp2->flags.roaming) 
							  rc=access_network(c->code,ROAMING_ACCESS|REG_ACCESS|DHCP_ACCESS);
						else
								rc=access_network(c->code,REG_ACCESS|DHCP_ACCESS);
					} else
							rc=access_network(c->code,DHCP_ACCESS);
					if(!rc) {
						if (debug > 3)
                			report(LOG_INFO, "  dynamic IP address client requested is disallowed by access policy");
                		send_dhcpnack(pkt,cl.cliid,"Dynamic IP address you requested is disallowed by access policy",cl.max_len,cl.cliid_explicit);
                		free_client_info(&cl);
                		return;
            		}
				}
			}

			/* locate ciaddr client wants to renew in our current leases */
			hp2 = hp;	/* save bootptab entry, we may need it again */
			exists = FALSE;
        	iphashcode=hash_HashFunction((u_char*)&pkt->bp_ciaddr.s_addr,4);
	        hp=(struct host *)hash_Lookup(ipboundhshtbl,iphashcode,ipcmp,(hash_datum*)&(pkt->bp_ciaddr.s_addr));

			if (hp) { /* we have a current lease for this ciaddr */
				/* make sure cliid in current binding == requester's cliid */
				if (hp->flags.cliid) { /* binding has cliid */
					if (!CLIID_EQ(hp->cliid,cl.cliid)) {
						if (debug > 2)
							report(LOG_INFO, "  requested IP address %s already leased to another client ID", inet_ntoa(pkt->bp_ciaddr));
						send_dhcpnack(pkt,cl.cliid,"IP Address leased to someone else",cl.max_len,cl.cliid_explicit);
						free_client_info(&cl);
						return;
					}
				} else {
					report(LOG_ERR,"  IP address %s presently bound, but binding has no cliid!", inet_ntoa(hp->iaddr));
					send_dhcpnack(pkt,cl.cliid,"Server error: requested IP address bound, but binding has no cliid",cl.max_len,cl.cliid_explicit);
					report(LOG_ERR, "  freeing binding");
					reuse_addr(hp);
					free_client_info(&cl);
					return;
				}
				if (!hp->flags.expire) {
					if (debug > 3 || log_expirations) {
						report(LOG_INFO, "  current binding was assigned via BootP (not DHCP), will not use");
						report(LOG_INFO, "  Freeing binding");
					}
					send_dhcpnack(pkt,cl.cliid,"Current binding was assigned via Bootp (not DHCP)",cl.max_len,cl.cliid_explicit);
					reuse_addr(hp);
					free_client_info(&cl);
					return;
				}
				if (hp->expire <= now) {
				/* It is possible that hp->expire >= now, meaning that the current binding has already
				   expired.  (That can happen if it expired by garbage collection hasn't happened yet to remove it.)
				   Let's be safe and discard that binding (I'm not sure it is safe to extend it if it happens
				   to be just an OFFER that was never accepted, instead of a real lease).  Note that we don't
				   send a NAK here....we continue processing the REQUEST, just as if the client is not presently
				   bound. */
					reuse_addr(hp);
					hp=NULL;
					if (debug > 3 || log_expirations) {
						report(LOG_INFO, "  current binding expired (awaiting garbage collection), will not use");
						report(LOG_INFO, "  Freeing binding, will proceed (treating client as unbound)");
					}
				}
			} else { 
				/* we can't just exit with error here as we'd like, since this could be a renew_unbound_statics case */
			}

			/* locate cliid in our current leases */
			hp3 =(struct host *)hash_Lookup(boundhshtbl,cl.clhashcode,boundcliidcmp,cl.cliid);
			if (hp3) { /* cliid found in boundhshtbl */
				if (!hp) { /* must check since we couldn't rule out this case earlier, due to renew_unbound_statics possibility */
					/* perhaps there are two clients trying to use same cliid */
					report(LOG_ERR, "  dhcp_request(): RENEWING/REBINDING/BOUND: did *NOT* find client's IP (%s) in ipboundhshtbl,...", inet_ntoa(pkt->bp_ciaddr));
					report(LOG_ERR, "  but did find entry for this cliid (%s) in boundhshtbl, assigned IP address %s!", cliid_btoa(cl.cliid), inet_ntoa(hp3->iaddr));
					report(LOG_ERR, "freeing binding");
					reuse_addr(hp3);
					/* don't NAK since may be REBINDING */
					free_client_info(&cl);
					return;
				}
				/* make sure ip address in current binding == requested IP */
				if (hp3->iaddr.s_addr == hp->iaddr.s_addr)  {
						/* this is normal, all matches */
						exists=TRUE;
				} else {
					/* not sure what this means.  cliid apparently bound a different IP.  Unbind cliid and ignore */
					report(LOG_ERR, "  dhcp_request(): RENEWING/REBINDING/BOUND: Found client's IP (%s) in ipboundhshtbl for this cliid...", inet_ntoa(pkt->bp_ciaddr));
					report_cliid(LOG_ERR, cl.cliid);
					report(LOG_ERR, "but entry for this cliid in ipboundhshtbl is for *DIFFERENT* IP address (%s)!", inet_ntoa(hp3->iaddr));
					report(LOG_ERR, "freeing both bindings");
					reuse_addr(hp);
					reuse_addr(hp3);
					/* don't NAK since may be REBINDING */
					free_client_info(&cl);
					return;
				}
			} else { /* cliid not found in boundhshtbl */
				if (hp) { /* requested IP addr found in ipboundhshtbl */
					/* this should never happen, since we already know that the ipboundhshtbl record specifies this cliid */
					/* not much we can do here, other than log the error, and ignore the request */
					report(LOG_ERR, "  dhcp_request(): RENEWING/REBINDING/BOUND: Found client's IP (%s) in ipboundhshtbl for this cliid...", inet_ntoa(pkt->bp_ciaddr));
					report(LOG_ERR, cl.cliid);
					report(LOG_ERR, "but did *NOT* find entry for this cliid in boundhshtbl!");
					report(LOG_ERR, "freeing binding");
					reuse_addr(hp);
					/* don't NAK since may be REBINDING */
					free_client_info(&cl);
					return;
				} else { /* requested IP addr not found in ipboundhshtbl */
					/* this may be normal; neither cliid nor req ip was bound, exists still == FALSE */
				}
			}

			if (!hp) { /* client not presently bound to us -- we have no current lease for this client */
				if (debug > 3)
					report(LOG_INFO, "  client's ciaddr %s is not presently bound", inet_ntoa(pkt->bp_ciaddr));
				if (!hp2->flags.dynamic && renew_unbound_statics) {
					/* According to spec, we should NOT renew, as we have no current lease for client.
					   However, if client is trying to renew an IP address that appears in our bootptab
					   as statically-assigned to this cliid (and the request has passed all validity checks
					   up to this point), and the server was started with the -u (renew_unbound_statics) option,
					   we will go ahead and issue a lease for this client anyway.  (Internally, we are creating
					   a new lease; from the client's point of view, we are renewing an existing lease.  So
					   this could certainly be a problem for the client if the client's idea of the lease include some info
					   that does not agree with the what we provide in the ACK.)  Note that this behavior violates
					   the spec, so it is not default.  It is provided for two reasons:  (1) to allow you to more
					   easily migrate from an older dhcp server code (we didn't inherit oustanding lease info from
					   it, but clients still think they are bound to us, and (2) to allow multiple DHCP servers
					   that serve the same static IP addresses to provide some redundancy for each other.  In
					   case (2), if the client is bound to a server that is temporarily unavailable, this means
					   that when the client enters REBINDING state, the other servers can RENEW the client's
					   lease.  
					   We don't provide this functionality for dynamically-assigned addresses, as for case (1)
					   that seems less safe, and for case (2) we must assume, in the absence of a server-to-server
					   protocol, that each server has an independant range of dynamic addresses to distribute.
					   Note that this functionality will be entirely superfluous once there is a server-to-server protocol. */
					hp = hp2;	/* we'll use the assignment provided by the bootptab */
					if (debug > 3)
						report(LOG_INFO, "  -u option specified, and ciaddr is in bootptab as static for this cliid; will use");
					assigning=TRUE;	/* we normally don't report assignments for RENEWAL, but renewal of unbound statics
									   is really a server assignment, so we report it */
					/* If we have a current binding for this cliid, clear it; client obviously not still using it */
					hp2=(struct host*)hash_Lookup(boundhshtbl,cl.clhashcode,boundcliidcmp ,cl.cliid);
					if (hp2) {
						if (debug > 3 || log_expirations) {
							report(LOG_INFO, "  cliid presently bound to %s", inet_ntoa(hp2->iaddr));
							report(LOG_INFO, "  freeing binding");
						}
						reuse_addr(hp2);
					}
				} else {
					/* if we knew for SURE the client was RENEWING, we'd like to NAK, but this
				   		client may be REBINDING, so the best we can do is be silent.  The RENEWING
				   		clients will get to keep using their invalid leases until they expire at the client end,
						unfortunately, but at least we won't break REBINDING clients. */
					/* Given the lack of a server-to-server protocol, each server that has a pool
				   		of dynamic IP addresses must have a separate such pool, which means that this
				   		server *is* all-knowing for its dynamic IP addresses.  If you make that assumption,
				   		you could then add code here to check if ciaddr's entry in bootptab has the dynamic
				   		flag set; if it does, then you can safely send a NAK instead of being silent.  Unfortunately,
				   		that will stop working once there is a server-to-server protocol so multiple servers can
				   		share the same pool of dynamic IP addresses.  So I haven't implemented that change here. */
					if (debug > 3) {
						report(LOG_INFO, "  want to NAK if client is RENEWING, but client may be REBINDING, so will ignore");
					}
					free_client_info(&cl);
					return;
				}
			}

			/* all ok, fall through to sending ACK below */

			} /* RENEWING/REBINDING/BOUND/confused */

		} /* (cl.serv_addr.s_addr==0) */
	
	/* Grant/renew lease 'hp' to client */

	if (!hp) { /* debugging code */
		report(LOG_ERR, "  Error! About to send DHCPACK, but hp==NULL!");
		free_client_info(&cl);
		return;
	}


	if (!exists) { /* it wasn't in the bound tables (or we removed it since it was a bootp binding) */
		if (debug > 1)
			report(LOG_INFO,"  Will grant new lease on IP address %s (%s)",inet_ntoa(hp->iaddr), hp->hostname->string);
		if (!(hp=create_stub(hp,&cl))) { /* create the new struct host for the current bindings */
            report(LOG_ERR, "   dhcp_request(): create_stub() returned NULL, cannot create new binding");
            /* reuse_addr(hp); we can't call this since hp is toast.  We leak the IP address if dynamic, but no memory */
            free_client_info(&cl);
            return;
		}

		hp->xid=pkt->bp_xid;
		hp->linkcount+=2; /* starts =0; about to insert into 2 hashtabs */
#ifdef DEBUG
		if (debug > 15)
			report(LOG_DEBUG, "  will insert into boundhshtbl...");
#endif
		if (!hash_Insert(boundhshtbl,cl.clhashcode, boundcliidcmp, cl.cliid,(hash_datum*)hp)) {
			leases++;
#ifdef DEBUG
			if (debug > 15)
				report(LOG_DEBUG, "  incremented leases counter to %d", leases);
#endif
		} else {
			report(LOG_ERR, "    dhcp_request(): hash_Insert(boundhshtbl) failed (cliid=%s)", cliid_btoa(cl.cliid));
			report(LOG_ERR, "    dhcp_request(): not increasing 'leases' counter...so count may no longer be right");
		}
		/* You can positively identify dhcp assignments in current bindings; they have hp->flags.expire==1.
		   You can then subdivide them as static (hp->code==RESERVED_FLAG) or dynamic (hp->code==DHCP_FLAG). */
#ifdef DEBUG
		if (debug > 15)
			report(LOG_DEBUG, "  will insert into ipboundhshtbl...");
#endif
		if (hash_Insert(ipboundhshtbl,hp->iphashcode, ipcmp,(hash_datum *)&(hp->iaddr.s_addr),(hash_datum*)hp)) {
			report(LOG_ERR, "    dhcp_request(): hash_Insert(ipboundhshtbl) failed for IP address %s (cliid=%s)", inet_ntoa(hp->iaddr), cliid_btoa(cl.cliid));
		}
	} else /* if (!exists) */
		if (debug > 1)
			report(LOG_INFO, "  Will renew existing lease on IP address %s (%s)",inet_ntoa(hp->iaddr), hp->hostname->string);

	/* we may have to do this even in exists==TRUE.  It *will* correctly update the hp in the current bindings */
	if (!hp->flags.complete) {  /* fill out more of the new struct host */
		if (!create_lease(hp,cl.cliid,cl.vend,cl.user,0)) {
           	report(LOG_ERR, "   dhcp_request(): create_lease() returned NULL, cannot create new binding");
			reuse_addr(hp);
           	free_client_info(&cl);
           	return;
		}
	}

	if(hp->flags.dynamic)
		hp->code=DHCP_FLAG;
	else
		hp->code=RESERVED_FLAG;
 
	hp->expire = now + hp->lease;
	hp->flags.expire = 1;

	/* Report if client left some MBZ flag bits set, and fix it in reply */
	if (ntohs(pkt->bp_flags) & BP_FLAGS_MBZ) {
		if (debug > 9)
			report(LOG_NOTICE, " some MBZ bits in flags field were set in DHCPREQUEST (cliid=%s)", cliid_btoa(cl.cliid));
		pkt->bp_flags = htons(ntohs(pkt->bp_flags) & ~ BP_FLAGS_MBZ);  /* clear MBZ bits for reply */
	}

    rpkt=(struct bootp *)send_buf;
    bcopy(pkt,rpkt,pktlen);
    rpkt->bp_op=BOOTREPLY;
    rpkt->bp_hops=0;
    rpkt->bp_secs=0;
    rpkt->bp_yiaddr.s_addr=hp->iaddr.s_addr;
    rpkt->bp_siaddr.s_addr = hp->bootserver.s_addr;
	if(fill_header(rpkt,hp,&bootsize)){
		free_client_info(&cl);
		reuse_addr(hp);
		report(LOG_ERR,"  fill_header() returned non-zero, huh?");
		return;
	}
    /* always stick our name in */
	/* note strncpy will right-pad with nul, clearing out any garbage - good */
	strncpy(rpkt->bp_sname, hostname, BP_SNAME_LEN);	/* may be overwritten if options field overflows */

	dest=hp->flags.reply_addr?hp->reply_addr.s_addr:0;


	opt.cur=0;
	opt.len=cl.max_len;
	insert_byte_opt(rpkt,&opt,TAG_DHCP_MSG,DHCPACK);
	insert_ulong_opt(rpkt,&opt,TAG_DHCP_SERVERID,my_ip_addr.s_addr);
	insert_ulong_opt(rpkt,&opt,TAG_DHCP_IPLEASE,htonl(hp->lease));
	if (compact_messages && debug) {
		char giaddrbuf[MAX_INETADDR_STRING_LEN];
		char iaddrbuf[MAX_INETADDR_STRING_LEN];

		strncpy(giaddrbuf, inet_ntoa(pkt->bp_giaddr), MAX_INETADDR_STRING_LEN-1);
		strncpy(iaddrbuf, inet_ntoa(hp->iaddr), MAX_INETADDR_STRING_LEN-1);
		report(LOG_INFO, "DHCPREQUEST  %s [%s]: lease %s (%s) until %s (%d seconds)",
			haddrtoa2(pkt->bp_chaddr, pkt->bp_hlen), giaddrbuf, iaddrbuf, hp->hostname->string, cctime(now+hp->lease), hp->lease);
	}
	if (debug > 9)
			report(LOG_INFO, "  Lease expires in %d seconds (%s)", hp->lease, cctime(now+hp->lease));
	if(hp->flags.renew) {
		insert_ulong_opt(rpkt,&opt,TAG_DHCP_RENEW,htonl(hp->renew));
		if (debug > 9) 
			report(LOG_INFO, "  Lease renew set for %d seconds (%s)", hp->renew, cctime(now+hp->renew));
	}
	if(hp->flags.rebind) {
		insert_ulong_opt(rpkt,&opt,TAG_DHCP_REBIND,htonl(hp->rebind));
		if (debug > 9) 
			report(LOG_INFO, "  Lease rebind set for %d seconds (%s)", hp->rebind, cctime(now+hp->rebind));
	}
	if(cl.user){
		insert_string_opt(rpkt,&opt,TAG_DHCP_USERCLASS,
					cl.user->data,cl.user->length);
	}

	/* fill in options requested by client's Parameter Request List option, if any */
    if (rc=do_req_opts(hp, cl.req_opts, (u_char *) rpkt, &opt, &sent_options)) {
		report(LOG_ERR,"  do_req_opts() returned %d, will not send DHCPACK to Client ID %s", rc, cliid_btoa(hp->cl));
		free_client_info(&cl);
		reuse_addr(hp);
		return;
	}

	if(hp->flags.not_rq) {
		/* bootptab entry has the NR tag, saying that our DHCP replies should include any values that appear
		   in the bootptab entry, even if they were "Not Requested" by the client in a Parm Request option and
		   are not values we always return */
		if (rc=do_opts(hp, (u_char *) rpkt, &opt, &sent_options)) { 
			report(LOG_ERR,"  do_opts() returned %d, will not send DHCPACK to Client ID %s", rc, cliid_btoa(hp->cl));
			free_client_info(&cl);
			reuse_addr(hp);
			return;
		}
	}

	insert_end_opt(rpkt,&opt);

	/* We're done building the DHCPACK, send it */
	if (debug > 12)
		report(LOG_INFO,"  Sending DHCPACK (max_len=%d, bytes left=%d)", opt.len,opt.len-opt.cur);
	else if (debug > 3)
		report(LOG_INFO,"  Sending DHCPACK");
	acks_sent++;
	len=300;
	if(opt.cur<VEND_OFFSET)
		len=cl.max_len;
	if(opt.cur>len)
		len=opt.cur;
	send_dhcp_reply(rpkt,len,dest);

	if (log_assignments && assigning)
		report_assignment(1, hp->flags.dynamic, inet_ntoa(hp->iaddr), hp->hostname->string, &cl);

	update_lastbind(&cl,hp);
	free_client_info(&cl);

}

void dhcp_release(pkt)
struct bootp *pkt;
{
	struct client_info cl;
	struct host *hp,*hp2;
	unsigned int32 dest; /* network byte order */
	unsigned iphashcode,hwhashcode,rc;

	bzero(&cl,sizeof(cl));
	cl.max_len=576;		/* for testing overload functions, temporarily reduce this, e.g. to 300 */

	if(debug>1)
		report(LOG_INFO," processing DHCPRELEASE");
	releases_received++;
	rc=dhcp_opts(pkt,&cl);

	/* we don't need msg here, dhcp_opts() already produces one
	if(!cl.cliid)
		report(LOG_ERR,"  couldn't determine client identifier, ignoring dhcprelease");
	*/
	if(rc){
		free_client_info(&cl);
		return;
	}

	if (compact_messages && debug)
		report(LOG_INFO, "DHCPRELEASE  %s [%s]: request", haddrtoa2(pkt->bp_chaddr, pkt->bp_hlen), inet_ntoa(pkt->bp_giaddr));

	if(debug>1)
		report_cliid(LOG_INFO,cl.cliid);

	if (cl.serv_addr.s_addr!=0 && !is_my_ip_addr(& cl.serv_addr)) {
		if(debug>3)
			report(LOG_INFO,"  message is not for me (Server Identifier option is %s), ignoring",inet_ntoa(cl.serv_addr));
		free_client_info(&cl);
		return;
	}

	cl.clhashcode=hash_HashFunction(cl.cliid->data,cl.cliid->length);
	hp =(struct host *)hash_Lookup(boundhshtbl,cl.clhashcode,boundcliidcmp,cl.cliid);
	if(!hp){
		if(debug>3)
			report(LOG_INFO,"  binding doesn't exist for this client id, ignoring");
		free_client_info(&cl);
		return;
	}

	if(cl.msg)
		if (debug > 3)
			report(LOG_INFO,"  client specified message option=%s",cl.msg);

	iphashcode=hash_HashFunction((u_char*)&hp->iaddr.s_addr,4);
	hp2 =(struct host *)hash_Lookup(ipboundhshtbl,iphashcode,ipcmp,(hash_datum *)&(hp->iaddr.s_addr));
	if (!hp2) {
		report(LOG_ERR, "dhcp_release(): entry for cliid (%s) found in boundhshtbl specifying IP address %s...", cliid_btoa(hp->cl), inet_ntoa(hp->iaddr));
		report(LOG_ERR, "but no entry for this IP addr found in ipboundhshtbl!");
		report(LOG_ERR, "will still free binding");
	} else {
		if (!hp2->flags.cliid) {
			report(LOG_ERR, "dhcp_release(): entry for cliid (%s) found in boundhshtbl specifying IP address %s...", cliid_btoa(hp->cl), inet_ntoa(hp->iaddr));
			report(LOG_ERR, "but ipboundhshtbl entry for this IP addr lacks a cliid!");
			report(LOG_ERR, "will still free both bindings");
		} else if (!BINDATA_EQ(hp2->cliid, cl.cliid)) {
			report(LOG_ERR, "dhcp_release(): entry for cliid (%s) found in boundhshtbl specifying IP address %s...", cliid_btoa(hp->cl), inet_ntoa(hp->iaddr));
			report(LOG_ERR, "but ipboundhshtbl entry for this IP addr has different cliid (%s)", cliid_btoa(hp2->cl));
			report(LOG_ERR, "will still free both bindings");
		}
	}

	if (compact_messages && debug) {
		char giaddrbuf[MAX_INETADDR_STRING_LEN];
		char iaddrbuf[MAX_INETADDR_STRING_LEN];

		strncpy(giaddrbuf, inet_ntoa(pkt->bp_giaddr), MAX_INETADDR_STRING_LEN-1);
		strncpy(iaddrbuf, inet_ntoa(hp->iaddr), MAX_INETADDR_STRING_LEN-1);
		report(LOG_INFO, "DHCPRELEASE  %s [%s]: release %s (%s)",
			haddrtoa2(pkt->bp_chaddr, pkt->bp_hlen), giaddrbuf, iaddrbuf, hp->hostname->string);
	}

	if(debug>2 || log_expirations)
		report(LOG_INFO,"  freeing binding to %s",inet_ntoa(hp->iaddr));
	reuse_addr(hp);

	free_client_info(&cl);

	return;
}

void dhcp_decline(pkt)
struct bootp *pkt;
{
	struct client_info cl;
	struct host *hp,*hp2;
	unsigned int32 dest;
	struct order_list **p;
	unsigned rc;

	bzero(&cl,sizeof(cl));
	cl.max_len=576; 	/* for testing overload functions, temporarily reduce this, e.g. to 300 */
	if(debug>1)
		report(LOG_INFO," processing DHCPDECLINE");
	declines_received++;
	rc=dhcp_opts(pkt,&cl);

	/* we don't need msg here, dhcp_opts() already produces one
	if(!cl.cliid)
		report(LOG_ERR,"  couldn't determine client identifier, ignoring dhcpdecline");
	*/
	if(rc){
		free_client_info(&cl);
		return;
	}

	if(debug>1)
		report_cliid(LOG_INFO,cl.cliid);
	cl.clhashcode=hash_HashFunction(cl.cliid->data,cl.cliid->length);
	hp =(struct host *)hash_Lookup(boundhshtbl,cl.clhashcode,boundcliidcmp,cl.cliid);

	if(hp) {
		if (compact_messages && debug) {
			char giaddrbuf[MAX_INETADDR_STRING_LEN];
			char iaddrbuf[MAX_INETADDR_STRING_LEN];

			strncpy(giaddrbuf, inet_ntoa(pkt->bp_giaddr), MAX_INETADDR_STRING_LEN-1);
			strncpy(iaddrbuf, inet_ntoa(hp->iaddr), MAX_INETADDR_STRING_LEN-1);
			report(LOG_INFO, "DHCPDECLINE  %s [%s]: freeing %s (%s)",
				haddrtoa2(pkt->bp_chaddr, pkt->bp_hlen), giaddrbuf, iaddrbuf, hp->hostname->string);
		}
		if (debug>1 || log_expirations)
			report(LOG_INFO,"  freeing binding for IP address %s",inet_ntoa(hp->iaddr));
		reuse_addr(hp);
	} else { /* hp */
		if (compact_messages && debug)
			report(LOG_INFO, "DHCPDECLINE  %s [%s]; binding not found",
				haddrtoa2(pkt->bp_chaddr, pkt->bp_hlen), inet_ntoa(pkt->bp_giaddr));
		if (debug > 1)
			report(LOG_INFO,"  no binding for this clientid found, ignoring");
	}

	free_client_info(&cl);

}

void dhcp_inform(pkt)
struct bootp *pkt;
{
	struct client_info cl;
	struct host *hp,*hp2,generic;
	struct opts opt;
	char override[2000];
	int len;
	unsigned rc,iphashcode,hwhashcode,nmhashcode;


	bzero(&cl,sizeof(cl));
	cl.max_len=576; 	/* for testing overload functions, temporarily reduce this, e.g. to 300 */
	if(debug>1)
		report(LOG_INFO," processing DHCPINFORM");
	informs_received++;
	rc=dhcp_opts(pkt,&cl);

	/* we don't need msg here, dhcp_opts() already produces one
	if(!cl.cliid)
		report(LOG_ERR,"  couldn't determine client identifier, ignoring dhcpinform");
	*/
	if(rc){
		free_client_info(&cl);
		return;
	}

	if (compact_messages && debug)
		report(LOG_INFO, "DHCPINFORM   %s [%s]: ignoring", haddrtoa2(pkt->bp_chaddr, pkt->bp_hlen), inet_ntoa(pkt->bp_giaddr));

	if(debug>1)
		report_cliid(LOG_INFO,cl.cliid);

	/* the entire dhcp_inform() function appears to be so screwed-up at this time,
   I'm temporarily replacing it with a no-op.  The function didn't appear to actually
   do anything resembling what the spec says it should do.  The response packet
   it sent seemed to be little more than an echo of the INFORM packet it got, with a
   few fields cleared or changed.  The response packet seemed to have little to do with
   the information obtained from the bootptab.  */

#if 0
	cl.clhashcode=hash_HashFunction(cl.cliid->data,cl.cliid->length);
	/* Look for clientid in bootptab.  (Remember that if incoming packet doesn't specify a 
	   DHCP Client ID option, we've computed its ClientID by concatenating its htype with its haddr.) */
	hp =(struct host *)hash_Lookup(clhashtable, cl.clhashcode,cliidcmp,cl.cliid);

#ifdef RETRY_HADDR_AS_CLIID
	/* I've ifdef'd out this because pretending to find a cliid match for clients when there's a mismatch 
	   between the packet's Cliid and the bootptab's cliid can cause problems later as we start storing
	   the inconsistent values in our hashtables.  Besides, we probably shouldn't be pretending to find
	   a match in the first place -- errors in the cliid specified by the bootptab or the client
	   should be corrected, not ignored.
	*/
	if(!hp){
		/* We didn't find ClientID in bootptab's clhashtable using ClientID as key.  
		   (Recall that clhashtable contains bootptab entries keyed by cliid (if 'cl' explicitly 
		   specified in bootptab entry) else bootptab entries keyed by haddr.) 
		   Perhaps the incoming packet contains a DHCP ClientID option with a garbage value.
		   Or perhaps the bootptab entry specifies a garbage cliid (via an incorrect 'cl' tag).
		   In either case, we may have more luck if we look up the incoming packet's haddr in
		   bootptab's hwhastable using haddr as key.  
		*/
		hwhashcode=hash_HashFunction(cl.cliid->data+1,cl.cliid->length-1);
		hp =(struct host *)hash_Lookup(hwhashtable, hwhashcode, cliidhwlookcmp,cl.cliid);
	}
#endif

	if (hp) {
		if (debug>3)
			report(LOG_INFO, "  found cliid in bootptab");
	} else {
		if (debug>3)
			report(LOG_INFO, "  did not find cliid in bootptab");
		iphashcode=hash_HashFunction((u_char*)&pkt->bp_ciaddr.s_addr,4);
		hp=(struct host *)hash_Lookup(iphashtable,iphashcode,ipcmp, &(pkt->bp_ciaddr.s_addr));
		if (debug>3)
			if (hp)
				report(LOG_INFO, "  found ciaddr (%s) in bootptab", inet_ntoa(pkt->bp_ciaddr));
			else
				report(LOG_INFO, "  did not find ciaddr (%s) in bootptab", inet_ntoa(pkt->bp_ciaddr));
	}

	if (hp) { /* client is registered */
		if (!hp->flags.iaddr) {
			hp=NULL;
			if (debug > 3)
				report(LOG_INFO, "  will ignore bootptab entry as it lacks an IP address");
		} else {
			if (!(hp=create_stub(hp, &cl))) {    /* create the new struct, fill in a few values */
				report(LOG_ERR, "   dhcp_inform(): create_stub() returned NULL, cannot create new stub");
				free_client_info(&cl);
				return;
			}
			/* we use create_lease() to get some inheritance */
			if (!create_lease(hp, cl.cliid, cl.vend, cl.user, 0)) { 
        		report(LOG_ERR, "   dhcp_inform(): create_lease() returned NULL, cannot build response");
				hp->linkcount = 1; /* to force free_host() to free it, as it starts =0 */
				free_host(hp);
        		free_client_info(&cl);
        		return;
			}
		}
	} else { /* client isn't registered, or bootptab entry lacks an IP address */
		/*  construct a generic client based on $C, $V, or $DHCP bootptab entries if possible */
		if (debug>3)
				report(LOG_INFO, "  will try to create generic response");
		if(cl.user){
			override[0]='$';
			override[1]='C';
			override[2]=0;
			for(i=0;i<cl.user->length;i++)
				sprintf(override+strlen(override),"%02X",cl.user->data[i]);
			if(debug>3)
				report(LOG_INFO, "  will look for $C bootptab override entry");
		} else if(cl.vend){
			override[0]='$';
			override[1]='V';
			override[2]=0;
			for(i=0;i<cl.vend->length;i++)
				sprintf(override+strlen(override),"%02X",cl.vend->data[i]);
			if(debug>3)
				report(LOG_INFO, "  will look for $V bootptab override entry");
		} else  {
			strcpy(override,"$DHCP");
			if(debug>3)
				report(LOG_INFO, "  will look for $DHCP bootptab override entry");
		} 
		bzero(&generic,sizeof(generic));
		nmhashcode=hash_HashFunction((u_char*)override,strlen(override));
		hp2=(struct host *)hash_Lookup(nmhashtable,nmhashcode,nmcmp,override);
		if(hp2) {
			if(debug>3)
				report(LOG_INFO, "  found bootptab override entry, will use that");
			fill_defaults2(&generic,hp2);	/* copies from hp2 into generic */
		} else {
			if(debug>3)
				report(LOG_INFO, "  no bootptab override entry found, ignoring client");
			free_client_info(&cl);
			return;
		}
	}


	/* At this point, either hp is set, based upon finding the client in the bootptab,
	   otherwise generic is set, based upon finding a template entry in the bootptab. */

	/* Remove lease time parameters, if any; those should not be in a response to a 
	   DHCPINFORM packet */
	if (hp)
		hp->flags.lease = hp->flags.renew = hp->flags.renew = 0;
	else
		generic.flags.lease = generic.flags.renew = generic.flags.renew = 0;

	/* Report if client left some MBZ flag bits set, and fix it in reply */
	if (ntohs(pkt->bp_flags) & BP_FLAGS_MBZ) {
		if (debug > 9) {
			report(LOG_NOTICE, " some MBZ bits in flags field were set in DHCPINFORM (cliid=%s)", cliid_btoa(cl.cliid));
		}
		pkt->bp_flags = htons(ntohs(pkt->bp_flags) & ~ BP_FLAGS_MBZ);  /* clear MBZ bits for reply */
	}

	/* Why does the response packet we build have little to do with the info we have
	   gotten into 'hp' or 'generic' ? */

	/* why do we malloc here instead of using send_buf ?*/
	rpkt=(struct bootp *)smalloc(cl.max_len);
	bcopy(pkt,rpkt,pktlen);
	rpkt->bp_op=BOOTREPLY;
	rpkt->bp_hops=0;
	rpkt->bp_secs=0;
	/* spec says we do NOT fill in yiaddr.  */
	/* rpkt->bp_yiaddr.s_addr=rpkt->bp_ciaddr.s_addr; */
	if (hp)
		rpkt->bp_siaddr.s_addr = hp->bootserver.s_addr;
	else
		rpkt->bp_siaddr.s_addr = generic.bootserver.s_addr;

	opt.cur=0;
	opt.len=cl.max_len;
	insert_byte_opt(rpkt,&opt,TAG_DHCP_MSG,DHCPACK);
	insert_ulong_opt(rpkt,&opt,TAG_DHCP_SERVERID,my_ip_addr.s_addr);
	if(cl.user){
		insert_string_opt(rpkt,&opt,TAG_DHCP_USERCLASS,
					cl.user->data,cl.user->length);
		del_bindata(cl.user);
	}
	if(cl.vend)
		del_bindata(cl.vend);
	insert_end_opt(rpkt,&opt);

	if (debug > 12)
		report(LOG_INFO,"  Sending DHCPACK (max_len=%d bytes left=%d)", opt.len,opt.len-opt.cur);
	else if (debug > 3)
		report(LOG_INFO,"  Sending DHCPACK");
	acks_sent++;
	len=300;
	if(opt.cur<VEND_OFFSET)
		len=cl.max_len;
	if(opt.cur>len)
		len=opt.cur;
	send_dhcp_reply(rpkt,len,rpkt->bp_ciaddr.s_addr);

	free(rpkt);

	free_client_info(&cl);

	if (hp) {
		hp->linkcount = 1; /* was 0, this will force it to be freed */
		free_host(hp);
	}

#else

	if(debug>1)
		report(LOG_INFO, "  dhcp_inform() not implemented, ignoring");

#endif
}



/* Macro for use in the function below */
/* adds a request for TAGNAME to req_opts[] if MEMBER's exists in hp
   but hasn't been inserted into the reply packet yet */
#define ADD_REQUEST(MEMBER,TAGNAME) do \
{ \
	if (hp->flags.MEMBER && ! f->MEMBER) \
		req_opts[i++] = TAGNAME ; \
} while (0)


/* Fill in vendor options field in DHCP reply packet for those items we know
about but have not yet been previously filled in.  Return 0 on success, nonzero
on any error inserting any option.
Avoid inserting any requested option that has already been inserted.   */

int do_opts(hp, bp, cur, f)
struct host *hp;		/* client info, e.g. from bindings */
u_char *bp;				/* reply packet we are building */
struct opts *cur;		/* records current location & length of the options data in reply packet */
struct flag *f;			/* indicates which options have *already* been inserted into reply packet */
{
	
	static u_char req_opts[256];	/* which options to try to insert */
	i = 1;				/* we start inserting at req_opts[1], NOT 0 */

	/* since do_req_opts() already does what we need, for a specific list of requested options,
	   we'll just build a list of which options we want it to try to insert, and then let do_req_opts()
	   do the rest of the work. 
	   We don't need to bother with any of the tags that are ALWAYS inserted by us in DHCP reply
	   packets (e.g. TAG_DHCP_SERVERID, TAG_DHCP_IPLEASE, TAG_DHCP_REBIND, TAG_DHCP_RENEW, TAG_DHCP_USERCLASS),
	   since we know those have already been inserted if defined */

	/* this function still needs to be written; for now, insert nothing */
	/* return 0; */

	/* these will be inserted into the reply packet in the order listed below, so
	   you may wish to list more important ones first.  Of course, these only get
	   inserted after the always-inserted and the Requested options are inserted, anwway. */

	ADD_REQUEST(gateway, TAG_GATEWAY);
	ADD_REQUEST(subnet_mask, TAG_SUBNET_MASK);
	ADD_REQUEST(domain_server, TAG_DOMAIN_SERVER);
	ADD_REQUEST(domain_name, TAG_DOMAIN_NAME);
	ADD_REQUEST(broadcast_addr, TAG_BROADCAST_ADDR);
	ADD_REQUEST(send_name, TAG_HOST_NAME);
	ADD_REQUEST(nis_server, TAG_NIS_SERVER);
	ADD_REQUEST(nis_domain, TAG_NIS_DOMAIN);
	ADD_REQUEST(root_path, TAG_ROOT_PATH);
	ADD_REQUEST(swap_server, TAG_SWAP_SERVER);
	ADD_REQUEST(exten_file, TAG_EXTEN_FILE);
	ADD_REQUEST(log_server, TAG_LOG_SERVER);
	ADD_REQUEST(ntp_server, TAG_NTP_SERVER);
	ADD_REQUEST(time_server, TAG_TIME_SERVER);
	ADD_REQUEST(time_offset, TAG_TIME_OFFSET);
	ADD_REQUEST(lpr_server, TAG_LPR_SERVER);
	ADD_REQUEST(node_type, TAG_NODE_TYPE);
	ADD_REQUEST(nb_scope, TAG_NB_SCOPE);
	ADD_REQUEST(nbns_server, TAG_NBNS_SERVER);
	ADD_REQUEST(name_server, TAG_NAME_SERVER);
	ADD_REQUEST(rlp_server, TAG_RLP_SERVER);
	ADD_REQUEST(cookie_server, TAG_COOKIE_SERVER);
	ADD_REQUEST(impress_server, TAG_IMPRESS_SERVER);
	ADD_REQUEST(bootsize, TAG_BOOT_SIZE);
	ADD_REQUEST(dump_file, TAG_DUMP_FILE);

	/* BUG: if hp->flags.generic is set, we also need to try to insert any
	   generic options (also being careful not to re-insert any that were already
	   specified by tagname).  This needs to be implemented. */

	req_opts[0] = i ;	/* put count of requested options in first byte */
	return do_req_opts(hp, req_opts, bp, cur, f);
}



/* Fill in vendor options field in DHCP reply packet for those items
requested by client in a Parameter Request List option.  Return 0 on success.
Avoid inserting any requested option that has already been inserted.   */

int do_req_opts(hp, req_opts, bp, cur, f)
struct host *hp;		/* client info, e.g. from bindings */
u_char *req_opts;		/* list of options in Parameter Request List specified by client */
						/* Note that first byte is the count of (one-byte) options that follow */
u_char *bp;				/* reply packet we are building */
struct opts *cur;		/* records current location & length of the options data in reply packet */
struct flag *f;			/* indicates which options have *already* been inserted into reply packet */
{
	int i,j,rc;

	if (!req_opts)
		return 0;

	for (i=1; i < (int) (req_opts[0]+1); i++) {
		rc=0;
		switch(req_opts[i]){                       
			case TAG_NODE_TYPE:
				if (!f->node_type && hp->flags.node_type) {
					f->node_type=1;
					rc=insert_byte_opt((struct bootp *)bp,cur,TAG_NODE_TYPE,hp->node_type);
				}
				break;
			/* This is already handled in dhcp_request()
			case TAG_DHCP_REBIND:
				if (!f->rebind && hp->flags.rebind) {
					f->rebind=1;
					rc=insert_ulong_opt((struct bootp *)bp,cur,TAG_DHCP_REBIND,htonl(hp->rebind));
				}
				break;
			*/
			/* This is already handled in dhcp_request()
			case TAG_DHCP_RENEW:
			if (!f->renew && hp->flags.renew) {
				f->renew=1;
				rc=insert_ulong_opt((struct bootp *)bp,cur,TAG_DHCP_RENEW,htonl(hp->renew));
			}
			break;
			*/
			case TAG_SUBNET_MASK:
				if (!f->subnet_mask && hp->flags.subnet_mask) {
					f->subnet_mask=1;
					rc=insert_ulong_opt((struct bootp *)bp,cur,TAG_SUBNET_MASK,hp->subnet_mask.s_addr);
				}
				break;
			case TAG_NB_SCOPE:
				if (!f->nb_scope && hp->flags.nb_scope) {
					f->nb_scope=1;
					rc=insert_str_opt((struct bootp *)bp,cur,TAG_NB_SCOPE,hp->nb_scope->string);
				}
				break;
			case TAG_DOMAIN_NAME:
				if (!f->domain_name) { 
					f->domain_name=1;
					if (hp->flags.domain_name)
						rc=insert_str_opt((struct bootp *)bp,cur,TAG_DOMAIN_NAME,hp->domain_name->string);
					else {
						for(j=0;hp->hostname->string[j] && hp->hostname->string[j]!='.';j++);
						rc=insert_str_opt((struct bootp *)bp,cur,TAG_DOMAIN_NAME,hp->hostname->string+j+1);
					}
				}
				break;
			case TAG_HOST_NAME:
				if (!f->name_switch) {
					f->name_switch=1;
					f->send_name=1;	/* I'm guessing this is useless duplicate info that tracks var above */
					rc=insert_str_opt((struct bootp *)bp,cur,TAG_HOST_NAME,hp->hostname->string);
				}
				break;
			case TAG_COOKIE_SERVER:
		    	if (!f->cookie_server && hp->flags.cookie_server) {
		    		f->cookie_server=1;
					rc=insert_addrs_opt((struct bootp *)bp,cur,TAG_COOKIE_SERVER,hp->cookie_server->addr,
						hp->cookie_server->addrcount);
				}
				break;
			case TAG_IMPRESS_SERVER:
		    	if (!f->impress_server && hp->flags.impress_server) {
		    		f->impress_server=1;
					rc=insert_addrs_opt((struct bootp *)bp,cur,TAG_IMPRESS_SERVER,hp->impress_server->addr,
						hp->impress_server->addrcount);
				}
				break;
			case TAG_LOG_SERVER:
		    	if (!f->log_server && hp->flags.log_server) {
		    		f->log_server=1;
					rc=insert_addrs_opt((struct bootp *)bp,cur,TAG_LOG_SERVER,hp->log_server->addr,
						hp->log_server->addrcount);
				}
				break;
			case TAG_LPR_SERVER:
		    	if (!f->lpr_server && hp->flags.lpr_server) {
		    		f->lpr_server=1;
					rc=insert_addrs_opt((struct bootp *)bp,cur,TAG_LPR_SERVER,hp->lpr_server->addr,
						hp->lpr_server->addrcount);
				}
				break;
			case TAG_NAME_SERVER:
		    	if (!f->name_server && hp->flags.name_server) {
		    		f->name_server=1;
					rc=insert_addrs_opt((struct bootp *)bp,cur,TAG_NAME_SERVER,hp->name_server->addr,
						hp->name_server->addrcount);
				}
				break;
			case TAG_RLP_SERVER:
		    	if (!f->rlp_server && hp->flags.rlp_server) {
		    		f->rlp_server=1;
					rc=insert_addrs_opt((struct bootp *)bp,cur,TAG_RLP_SERVER,hp->rlp_server->addr,
						hp->rlp_server->addrcount);
				}
				break;
			case TAG_TIME_SERVER:
		    	if (!f->time_server && hp->flags.time_server) {
		    		f->time_server=1;
					rc=insert_addrs_opt((struct bootp *)bp,cur,TAG_TIME_SERVER,hp->time_server->addr,
						hp->time_server->addrcount);
				}
				break;
			case TAG_NIS_SERVER:
		    	if (!f->nis_server && hp->flags.nis_server) {
		    		f->nis_server=1;
					rc=insert_addrs_opt((struct bootp *)bp,cur,TAG_NIS_SERVER,hp->nis_server->addr,
						hp->nis_server->addrcount);
				}
				break;
			case TAG_NTP_SERVER:
		    	if (!f->ntp_server && hp->flags.ntp_server) {
		    		f->ntp_server=1; 
					rc=insert_addrs_opt((struct bootp *)bp,cur,TAG_NTP_SERVER,hp->ntp_server->addr,
						hp->ntp_server->addrcount);
				}
				break;
			case TAG_GATEWAY:
		    	if (!f->gateway && hp->flags.gateway) {
					f->gateway=1;
					rc=insert_addrs_opt((struct bootp *)bp,cur,TAG_GATEWAY,hp->gateway->addr,
						hp->gateway->addrcount);
				}
				break;
			case TAG_DOMAIN_SERVER:
		    	if (!f->domain_server && hp->flags.domain_server) {
		    		f->domain_server=1;
					rc=insert_addrs_opt((struct bootp *)bp,cur,TAG_DOMAIN_SERVER,hp->domain_server->addr,
						hp->domain_server->addrcount);
				}
				break;
			case TAG_NBNS_SERVER: 
		    	if (!f->nbns_server && hp->flags.nbns_server) {
		    		f->nbns_server=1;
					rc=insert_addrs_opt((struct bootp *)bp,cur,TAG_NBNS_SERVER,hp->nbns_server->addr,
						hp->nbns_server->addrcount);
				}
				break;
			case TAG_BROADCAST_ADDR:
				if (!f->broadcast_addr && hp->flags.broadcast_addr) {
					f->broadcast_addr=1;
					rc=insert_ulong_opt((struct bootp *)bp,cur,TAG_BROADCAST_ADDR,hp->broadcast_addr.s_addr);
				}
				break;
			case TAG_SWAP_SERVER:
				if (!f->swap_server && hp->flags.swap_server) {
					f->swap_server=1;
					rc=insert_ulong_opt((struct bootp *)bp,cur,TAG_SWAP_SERVER,hp->swap_server.s_addr);
				}
				break;
			case TAG_BOOT_SIZE: /* BUG - what if bootptab specify bootsize auto option - see how bootp handles this */
				if (!f->bootsize && hp->flags.bootsize) {
					f->bootsize=1;
					/* we've been carrying hp->bootsize as a uint32, but the option is really a uint16.  So before
					   sending it, lop off the high bytes and cast it.  */
					rc=insert_ushort_opt((struct bootp *)bp,cur,TAG_BOOT_SIZE, htons( (unsigned short) (hp->bootsize & 0xFFFFFFFF)));
				}
				break;
			case TAG_TIME_OFFSET: 
				if (!f->time_offset && hp->flags.time_offset) {
					f->time_offset=1;
					rc=insert_ulong_opt((struct bootp *)bp,cur,TAG_TIME_OFFSET,htonl(hp->time_offset));
				}
				break;
			case TAG_DUMP_FILE:
				if (!f->dump_file && hp->flags.dump_file) {
					f->dump_file=1;
					rc=insert_str_opt((struct bootp *)bp,cur,TAG_DUMP_FILE,hp->dump_file->string);
				}
				break;
			case TAG_EXTEN_FILE:
				if (!f->exten_file && hp->flags.exten_file) {
					f->exten_file=1;
					rc=insert_str_opt((struct bootp *)bp,cur,TAG_EXTEN_FILE,hp->exten_file->string);
				}
				break;
			case TAG_ROOT_PATH:
				if (!f->root_path && hp->flags.root_path) {
					f->root_path=1;
					rc=insert_str_opt((struct bootp *)bp,cur,TAG_ROOT_PATH,hp->root_path->string);
				}
				break;
			case TAG_NIS_DOMAIN:
				if (!f->nis_domain && hp->flags.nis_domain) {
					f->nis_domain=1;
					rc=insert_str_opt((struct bootp *)bp,cur,TAG_NIS_DOMAIN,hp->nis_domain->string);
				}
				break;
		/* BUG: we should also see if the requested option matches any of the generic-specified
		   options (if hp->flags.generic); if found, and this option has not already been inserted
		   (e.g. by name), we should insert it.  This needs to be implemented. */
			case TAG_DHCP_SERVERID:
			case TAG_DHCP_IPLEASE:
			case TAG_DHCP_REBIND:
			case TAG_DHCP_RENEW:
			case TAG_DHCP_USERCLASS:
			case TAG_VENDOR:
				/* These options are ones that we handle elsewhere, either because they cause
			   	values to be inserted elsewhere (not in the options field), or because they
			   	request options that we will always return in DHCPACK even if not requested. 
			   	So we simply ignore them if requested here. */
				break;
			default:
				if (debug > 9)
					report(LOG_INFO,"  ignoring unknown or unexpected option (%d) in client's Parameter Request List option", req_opts[i]);
				break;
		}
		if (rc) 
			return rc;	/* error inserting some option, give up */
	}

	return 0;
}

void send_dhcpnack(pkt,cliid,msg,max_len,cliid_explicit)
struct bootp *pkt;
struct shared_bindata *cliid;
char *msg;
unsigned int max_len;
boolean cliid_explicit;
{
	unsigned char *opts;
	int32 len;
	struct opts opt;
	struct bootp *tmp;

	tmp=(struct bootp*)send_buf;
	bcopy(pkt,tmp,pktlen);

	tmp->bp_op=BOOTREPLY;
	tmp->bp_hops=0;
	tmp->bp_secs=0;
	tmp->bp_ciaddr.s_addr=0;
	tmp->bp_siaddr.s_addr=0;
	tmp->bp_yiaddr.s_addr=0;
	tmp->bp_flags|=0x8000;

	opt.cur=0;
	opt.len=max_len;
	insert_byte_opt(tmp,&opt,TAG_DHCP_MSG,DHCPNACK);
	insert_ulong_opt(tmp,&opt,TAG_DHCP_SERVERID,my_ip_addr.s_addr);
	if (cliid_explicit)
		insert_string_opt(tmp,&opt,TAG_DHCP_CLIID,cliid->data,cliid->length);
	insert_str_opt(tmp,&opt,TAG_MESSAGE,msg);
	insert_end_opt(tmp,&opt);
	if (debug>1)
		report(LOG_INFO,"Sending DHCPNACK: %s",msg);
	nacks_sent++;

	len=300;
	if(opt.cur<VEND_OFFSET)
		len=max_len;
	if(opt.cur>len)
		len=opt.cur;
	send_dhcp_reply(tmp,len,0);
}

		

void send_dhcp_reply(pkt,send_len,dest)
struct bootp *pkt;
int32 send_len;
unsigned int32 dest; /* network byte order */
{
	struct sockaddr_in dest_addr;
    unsigned char *ha;
    int len, haf;
    struct in_addr dst;

	dest_addr.sin_family = AF_INET;
	dest_addr.sin_port = htons(bootpc_port);


	/* Here's our strategy for determine where to send the packet:

		If we were explicitly told where to send the packet, use that.

		Else if ciaddr is set, and broadcast bit is off, send to ciaddr.  
		(Ignoring the broadcast bit test, that's as per RFC; a client that fills
		in ciaddr must be prepared to accept a packet sent there.  And if the client's
		not on ciaddr's network, it is understood that the reply will never reach it.
		I added the broadcast bit test.)
		
		Else if giaddr is set, send to giaddr, changing the dstPort to bootpserver.
		
		Else if yiaddr==0, send to broadcast.  We understand that this won't reach client
		if client was non-local and unicast the request to us (preventing giaddr from
		being set).  We understand that this won't reach client if client was local
		AND we're multihomed AND the OS chooses to send the broadcast out an interface
		to which the client is not attached.  (We could try to fix that by trying to 
		detect this case, and then sending to a directed broadcast address to force it
		out a particular interface.  But that may not be accepted by some bootp clients,
		since bootp spec says that bootp broadcast must be sent to 255.255.255.255.)
		
		Else if yiaddr is non-local, send to yiaddr.  We understand that the test for
		non-local may not return an appropriate value if the server is multihomed.
		
		Else if broadcast bit is set, send to broadcast.
		
		Else send to yiaddr, priming the arp cache
	*/

	if (dest) {
		dest_addr.sin_addr.s_addr=dest;
		if (debug > 6)
			report(LOG_INFO, "  sending reply to specified reply address override: %s",
				inet_ntoa(dest_addr.sin_addr));
	} else if ((pkt->bp_ciaddr.s_addr) && !(ntohs(pkt->bp_flags) & 0x8000)){
		dest_addr.sin_addr.s_addr=pkt->bp_ciaddr.s_addr;
		if (debug > 6)
			report(LOG_INFO, "  sending reply to client-specified ciaddr %s",
				inet_ntoa(dest_addr.sin_addr));
	} else if (pkt->bp_giaddr.s_addr) {
		dest_addr.sin_addr.s_addr=pkt->bp_giaddr.s_addr;
		dest_addr.sin_port=htons(bootps_port);
		if (debug > 6)
			report(LOG_INFO, "  sending reply to giaddr: %s",
				inet_ntoa(dest_addr.sin_addr));
	} else if (! pkt->bp_yiaddr.s_addr) {
		dest_addr.sin_addr.s_addr = INADDR_BROADCAST;
		if (debug > 6)
			report(LOG_INFO, "  yiaddr==0, sending reply to broadcast");
	} else if (! is_local(& pkt->bp_yiaddr)) {
		dest_addr.sin_addr.s_addr = pkt->bp_yiaddr.s_addr;
		if (debug > 6)
			report(LOG_INFO, "  yiaddr is non-local, sending reply there: %s",
				inet_ntoa(dest_addr.sin_addr));
	} else if (ntohs(pkt->bp_flags) & 0x8000) {
		dest_addr.sin_addr.s_addr = INADDR_BROADCAST;
		if (debug > 6)
			report(LOG_INFO, "  yiaddr (%s) is local and bootp broadcast bit is set, sending reply to broadcast",
			inet_ntoa(pkt->bp_yiaddr));
	} else {
		if (debug > 6)
			report(LOG_INFO, "  yiaddr is local, sending reply there: %s", inet_ntoa(pkt->bp_yiaddr));
		dst = pkt->bp_yiaddr;
		ha = pkt->bp_chaddr;
		len = pkt->bp_hlen;
		if (len > MAXHADDRLEN)
	    	len = MAXHADDRLEN;
		haf = (int) pkt->bp_htype;
		if (haf == 0)
	    	haf = HTYPE_ETHERNET;
		if (debug > 6)
	    	report(LOG_INFO, "  setarp %s - %s", inet_ntoa(dst), haddrtoa(ha, len));
		setarp(s, &dst, haf, ha, len);
		dest_addr.sin_addr=dst;
	}


	if(sendto(s,(char*)pkt,send_len,0,(struct sockaddr *)&dest_addr,sizeof(dest_addr)) <0)
		 report(LOG_ERR, "  sendto: %s", get_network_errmsg());
}



void dhcp_garbage_collect()
{
 unsigned hashcode,iphashcode;
 struct host *hp,*hp2;
 struct order_list *t;
 struct cable *c;
 struct lastbind *oldbinding, tempoldbinding;


	if(debug>10)
		report(LOG_INFO,"DHCP garbage collect begin");

	for (hp=(struct host*)hash_FirstEntry(boundhshtbl); hp; hp=(struct host*)hash_NextEntry(boundhshtbl)) {

#ifdef DEBUG
		if (debug > 15) {
			if (! hp->cliid) {
				report(LOG_ERR, "dhcp_garbage_collect(): entry's hp->cliid is NULL, skipping!");
				continue;
			}
			report_cliid(LOG_DEBUG, hp->cliid);
		}
#endif

		if (!hp->flags.expire)  { /* lease is infinite */
			if (expire_infinite_dynamics) {	/* we want to expire it anyway, after some time */
				tempoldbinding.cliid = hp->cl;
				hashcode = hash_HashFunction(hp->cl->data,hp->cl->length);
				oldbinding = (struct lastbind*)hash_Lookup(lastbindhshtbl,hashcode,lastbindcmp,&tempoldbinding);
				if (!oldbinding) {
					report(LOG_ERR, "dhcp_garbage_collect(): entry in boundhshtbl has infinite lease, but no entry in lastbindhshtbl!");
					report(LOG_ERR, "...something is very wrong!");
					continue;	/* not much we can do */
				}
				if (now > oldbinding->last_touch + expire_infinite_dynamics) {
					if (debug>11 || log_expirations)
						report(LOG_INFO, "%s is infinite, but older than expire_infinite_dynamics; will remove binding", inet_ntoa(hp->iaddr));
					/* should we increment ips_expired here ? */
					reuse_addr(hp); 
					/* note reuse_addr() also sets hp->expire to NOW in the lastbinding entry.  That's good, as it means
					   that the next time we start, we won't re-instate the binding, even if expire_infinite_dynamics is unset at
					   that time.  But it does mean that examining the lastbind info will show the entry had a non-infinite
					   lease, which (for static addresses) makes them appear to have been assigned via DHCP instead of BOOTP.
					   That's confusing, but not a real problem, since we're not going to re-instate them when restarting
					   anyway, as they have expired. */
 
				} else {
					if (debug>11)
						report(LOG_INFO, "%s is infinite, and newer than expire_infinite_dynamics; retained", inet_ntoa(hp->iaddr));
				}
				continue;
			} else {	/* we never expire infinite leases */
#ifdef DEBUG
				if (debug > 15)
					report(LOG_DEBUG, " %s is infinite; retained", inet_ntoa(hp->iaddr));
#endif
				continue;
			}
		} else { /* lease is finite */

			if (hp->expire > now) {
#ifdef DEBUG
				if (debug > 15)
					report(LOG_DEBUG, " %s is finite and hasn't expired; retained", inet_ntoa(hp->iaddr));
#endif
				continue; /* keep lease, so proceed to next entry in bound table */
			} else {
				ips_expired++;
				if(debug>11 || log_expirations)
					report(LOG_INFO," %s is finite and has expired, will remove binding",inet_ntoa(hp->iaddr));
				reuse_addr(hp);
			}
 		}

	} /* walk through bound table */


	if(debug>10)
		report(LOG_INFO,"DHCP garbage collect end");

	now=time(0);
	dhcp_alarm=now+gcinterval;
}



/* Given a struct host * hp (a record that is a current binding, or intended to become a current
binding), modify it to inherit appropriate information from the bootptab.  First take into account any 
$C, $V, or $DHCP override entries in the bootptab.   Then take into account the bootptab entry for
hp's IP address.  We make *sure* the entry has a lease time, limiting it to no more than 'secs' if specified.
We also make *sure* the entry has a renew time and a rebind time, setting each to the default fractions
of the lease time (I'm not sure we really need to do this).
In addition to modifying it, we return it; return NULL on error. */

struct host *create_lease(hp,cliid,vend,user,secs)
struct host *hp;	/* the struct host we modify */
struct shared_bindata *cliid,*vend,*user;
unsigned secs;		/* if nonzero, ensure the lease is no more than this */
{
	struct host *hp2;
	unsigned hashcode;
	char override[2000];
	

	/* If request packet contained user class info option, vendor specific info option, else is DHCP, take into
	   account $C, $V, or $DHCP entry in bootptab.  Do this by setting 'override' to the name of the bootptab
	   entry we should look for */
	if(user){ /* request packet contains a user class info option */
		override[0]='$';
		override[1]='C';
		override[2]=0;
		for(ui=0;ui<user->length;ui++)
			sprintf(override+strlen(override),"%02X",user->data[ui]);
	} else if(vend){ /* request packet contains a vendor specific info option */
		override[0]='$';
		override[1]='V';
		override[2]=0;
		for(ui=0;ui<vend->length;ui++)
			sprintf(override+strlen(override),"%02X",vend->data[ui]);
	} else  /* request packet contains neither of above, but is dhcp */
		strcpy(override,"$DHCP");

	/* look for bootptab entry with name that matches value of 'override' */
	hashcode=hash_HashFunction((u_char*)override,strlen(override));
	hp2=(struct host *)hash_Lookup(nmhashtable,hashcode,nmcmp,override);
	if(hp2) /* there was an entry with this name! */
		fill_defaults2(hp,hp2); /* modify hp; info from hp2 is used to override info in hp */


	/* Locate the bootptab entry corresponding to hp's IP address, if any */
	if (hp->flags.iaddr) {
		hp2=(struct host*)hash_Lookup(iphashtable,hp->iphashcode,ipcmp, &(hp->iaddr.s_addr));
		if(!hp2)
			return NULL;	/* BUG: is that right? */
		fill_defaults2(hp,hp2); /* modify hp; info from hp2 is used to override info in hp */
	} else {
		/* BUG: should we try to find a bootptab entry corresponding to hp's hw address/cliid ?? */
		report(LOG_ERR, "create_lease(): called with hp where hp->flags.iaddr is FALSE");
		return NULL;
	}
	

	if(!hp->flags.lease){	/* ensure entry has a lease time */
		/* this only happens if we didn't fill in lease length via the steps above */
		hp->flags.lease=TRUE;
		hp->lease=DEFAULT_LEASE;
	}

	if(secs){	/* caller wants lease limited to no more than this */
		if(hp->flags.max_lease)
			if(secs<hp->flags.max_lease)
				hp->lease=secs;
		else
			if(secs<hp->lease)
				hp->lease=secs;
	}

	if(!hp->flags.renew){ /* ensure the entry has a renew time; I'm not sure this is really necessary */
		/* this only happens if we didn't fill in renew length via the steps above */
		hp->renew=hp->lease/2; 	/* we're using the default percentage from rfc */
		hp->flags.renew=TRUE;
	}
	if(!hp->flags.rebind){ /* ensure the entry has a rebind time; I'm not sure this is really necessary */
		/* this only happens if we didn't fill in rebind length via the steps above */
		hp->rebind=hp->lease*0.85;	/* we're using the default percentage from rfc */
		hp->flags.rebind=TRUE;
	}
		
	hp->flags.complete=TRUE; /* this flag seems to indicate that create_lease() has been used to 'complete'
							    this entry. */
	return hp;
}


/* Create a new struct host.  We read 'hp' as template to copy a few of the most critical items;
hp should contain hp->iaddr, hp->hostname, hp_iphashcode.
We'll actually use 'cl' (increasing its linkcount).
We return a ptr to an entirely new struct host we malloc. This is the primary way to create new struct hosts
that have to stick around (i.e. not just for temporary storage within a routine). 
It is the responsibility of the caller to set hp->linkcount to 1 (or more).
When you are done with it, call free_host() to destroy it. */

struct host *create_stub(hp,cl)
struct host *hp;		/* template host */
struct client_info *cl;	/* cl to link in */
{
	struct host *t;		

	if (! hp->flags.iaddr) {
		report(LOG_ERR, "create_stub(): called with hp where hp->flags.iaddr is FALSE");
		return NULL;
	}

	t=(struct host*)smalloc(sizeof(struct host));  /* that is where we build the one to return */
	bzero(t,sizeof(struct host));

	t->linkcount=0; 	/* yes, zero not one; it gets incremented later by the caller */

	t->flags.iaddr=TRUE;
	t->iaddr.s_addr=hp->iaddr.s_addr;

	t->hostname=hp->hostname;	/* note that we link in this shared string, not copy it */
	t->hostname->linkcount++;

	if(hp->flags.reply_addr) {
		t->reply_addr.s_addr=hp->reply_addr.s_addr;
		t->flags.reply_addr=TRUE; 
	}

	t->cl=cl->cliid;		/* note we link it in, not copy it */
	t->cl->linkcount+=2;	/* to account for the link we just made, and the one we're just about to */
	t->cliid = cl->cliid;
	t->flags.cliid=TRUE;

	t->iphashcode=hp->iphashcode;

	return t;
}


/* write an ASSIGNED report showing the vitals of an IP address assignment */
void
report_assignment(is_dhcp, is_dynamic, ipaddr, ipaddr_name, cl)
int is_dhcp;		/* 0=BOOTP 1=DHCP */
int is_dynamic;		/* 0=STATIC, 1=DYNAMIC */
char * ipaddr;		/* ip address being assigned */
char * ipaddr_name;	/* bootptab hostname for ip address being assigned */
struct client_info *cl;	/* from request packet */
{
	char message[2*MAXSTRINGLEN]; 
	struct host *hp;

	/* translate cliid to bootptab entry name (hostname) */
	/* Look for clientid in bootptab.  (Remember that if incoming packet doesn't specify a 
	   DHCP Client ID option, we've computed its ClientID by concatenating its htype with its haddr.) */
	hp =(struct host *)hash_Lookup(clhashtable, cl->clhashcode,cliidcmp,cl->cliid);

#ifdef RETRY_HADDR_AS_CLIID
	/* I've ifdef'd out this because pretending to find a cliid match for clients when there's a mismatch 
	   between the packet's Cliid and the bootptab's cliid can cause problems later as we start storing
	   the inconsistent values in our hashtables.  Besides, we probably shouldn't be pretending to find
	   a match in the first place -- errors in the cliid specified by the bootptab or the client
	   should be corrected, not ignored.
	*/
	if(!hp){ 
		/* We didn't find ClientID in bootptab's clhashtable using ClientID as key.  
		   (Recall that clhashtable contains bootptab entries keyed by cliid (if 'cl' explicitly 
		   specified in bootptab entry) else bootptab entries keyed by haddr.) 
		   Perhaps the incoming packet contains a DHCP ClientID option with a garbage value.
		   Or perhaps the bootptab entry specifies a garbage cliid (via an incorrect 'cl' tag).
		   In either case, we may have more luck if we look up the incoming packet's haddr in
		   bootptab's hwhastable using haddr as key.  
		*/
		unsigned hwhashcode;
		hwhashcode=hash_HashFunction(cl->cliidhw->data+1,cl->cliidhw->length-1);
		hp =(struct host *)hash_Lookup(hwhashtable, hwhashcode,cliidhwlookcmp,cl->cliidhw);
	}
#endif

	sprintf(message, "ASSIGNED %s %s %s (%s) TO CLIENT (%s) %s",
		(is_dhcp ? "DHCP" : "BOOTP"),
		(is_dynamic ? "DYNAMIC" : "STATIC"),
		ipaddr,
		ipaddr_name,
		(hp ? hp->hostname->string : ""),
		cliid_btoa(cl->cliid));
	report(LOG_INFO, message);
	return;
}
