/***********************************************************
        Copyright 1996 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <net/if.h>
#include <net/route.h>
#include <netdb.h>
#include <syslog.h>
#include <errno.h>


#include "defs.h"
#include "bootp.h"
#include "bootpd.h"
#include "dhcpd.h"
#include "dhcp.h"
#include "utils.h"


int insert_byte_opt(bp,opt,tag,value)
struct bootp *bp;
struct opts *opt;
unsigned char tag;
unsigned char value;
{
	int i;
	unsigned char *cp;

	cp=(unsigned char*)bp;
	if(opt->cur==0)
		opt->cur=VEND_OFFSET;
	
	if(opt->cur+7>opt->len){
		cp[opt->cur++]=TAG_DHCP_OVL;
		cp[opt->cur++]=1;
		cp[opt->cur++]=1; /* file overload */
		cp[opt->cur]=TAG_END;
		opt->cur=FILE_OFFSET;
	} else if (opt->cur<SNAME_OFFSET &&opt->cur+7>SNAME_OFFSET) {
		cp[opt->cur]=TAG_END;
		for(i=4;(u_long)bp->bp_vend[i]!=TAG_DHCP_OVL;i++);
		bp->bp_vend[i+2]=3;
		opt->cur=SNAME_OFFSET;
	} else if(opt->cur<VEND_OFFSET && opt->cur+7>VEND_OFFSET)
		return -1;
	cp[opt->cur++]=tag;
	cp[opt->cur++]=1;
	cp[opt->cur++]=value;

	return 0;
}

/* Insert 1-byte 'tag' and 4-byte 'value' into the packet represented by 'bp' and 'opt'.
   We try to put it in 'opt' if it fits, else overload it into an unused 'bp' header field.
   Note that 'value' is copied in byte-for-byte; if any network-byte ordering conversion needs
   to be done, it must be done before calling us */
int insert_ulong_opt(bp,opt,tag,value)
struct bootp *bp;
struct opts *opt;
unsigned char tag;
unsigned long value;
{
	int i;
	unsigned char *cp;

	cp=(u_char*)bp;
	if(opt->cur==0)
		opt->cur=VEND_OFFSET;
	
	if(opt->cur+11>opt->len){
		cp[opt->cur++]=TAG_DHCP_OVL;
		cp[opt->cur++]=1;
		cp[opt->cur++]=1; /* file overload */
		cp[opt->cur]=TAG_END;
		opt->cur=FILE_OFFSET;
	} else if (opt->cur+11>SNAME_OFFSET && opt->cur<SNAME_OFFSET){
		cp[opt->cur]=TAG_END;
		for(i=4;(u_long)bp->bp_vend[i]!=TAG_DHCP_OVL;i++);
		bp->bp_vend[i+2]=3;
		opt->cur=SNAME_OFFSET;
	} else if(opt->cur<VEND_OFFSET && opt->cur+11>VEND_OFFSET)
		return -1;
	cp[opt->cur++]=tag;
	cp[opt->cur++]=4;
	bcopy(&value,cp+opt->cur,4);
	opt->cur+=4;

	return 0;
}

/* Insert 1-byte 'tag' and 2-byte 'value' into the packet represented by 'bp' and 'opt'.
   We try to put it in 'opt' if it fits, else overload it into an unused 'bp' header field.
   Note that 'value' is copied in byte-for-byte; if any network-byte ordering conversion needs
   to be done, it must be done before calling us */
int insert_ushort_opt(bp,opt,tag,value)
struct bootp *bp;
struct opts *opt;
unsigned char tag;
unsigned short value;
{
	int i;
	unsigned char *cp;

	cp=(u_char*)bp;
	if(opt->cur==0)
		opt->cur=VEND_OFFSET;
	
	if(opt->cur+9>opt->len){
		cp[opt->cur++]=TAG_DHCP_OVL;
		cp[opt->cur++]=1;
		cp[opt->cur++]=1; /* file overload */
		cp[opt->cur]=TAG_END;
		opt->cur=FILE_OFFSET;
	} else if (opt->cur+9>SNAME_OFFSET && opt->cur<SNAME_OFFSET){
		cp[opt->cur]=TAG_END;
		for(i=4;(u_long)bp->bp_vend[i]!=TAG_DHCP_OVL;i++);
		bp->bp_vend[i+2]=3;
		opt->cur=SNAME_OFFSET;
	} else if(opt->cur<VEND_OFFSET && opt->cur+9>VEND_OFFSET)
		return -1;
	cp[opt->cur++]=tag;
	cp[opt->cur++]=2;
	bcopy(&value,cp+opt->cur,2);
	opt->cur+=2;

	return 0;
}

int insert_addrs_opt(bp,opt,tag,value,len)
struct bootp *bp;
struct opts *opt;
unsigned char tag;
struct in_addr *value;
int len;
{
	int i;
	unsigned char *cp;
	int bytes;

	cp=(u_char*)bp;
	if(opt->cur==0)
		opt->cur=VEND_OFFSET;
	bytes=len*sizeof(struct in_addr);	
	if(opt->cur+6+bytes>opt->len){
		cp[opt->cur++]=TAG_DHCP_OVL;
		cp[opt->cur++]=1;
		cp[opt->cur++]=1; /* file overload */
		cp[opt->cur]=TAG_END;
		opt->cur=FILE_OFFSET;
	} else if (opt->cur+6+bytes>SNAME_OFFSET && opt->cur<SNAME_OFFSET){
		cp[opt->cur]=TAG_END;
		for(i=4;(u_long)bp->bp_vend[i]!=TAG_DHCP_OVL;i++);
		bp->bp_vend[i+2]=3;
		opt->cur=SNAME_OFFSET;
	} else if(opt->cur<VEND_OFFSET && opt->cur+6+bytes>VEND_OFFSET)
		return -1;
	cp[opt->cur++]=tag;
	cp[opt->cur++]=bytes;
	bcopy(value,cp+opt->cur,bytes);
	opt->cur+=bytes;

	return 0;
}

int insert_end_opt(bp,opt)
struct bootp *bp;
struct opts *opt;
{
	unsigned char *cp;
	cp=(u_char*)bp;
	cp[opt->cur++]=TAG_END;
	return 0;
}
int insert_string_opt(bp,opt,tag,value,len)
struct bootp *bp;
struct opts *opt;
unsigned char tag;
char *value;
int len;
{
	int i;
	unsigned char *cp;
	int bytes;

	cp=(u_char*)bp;
	if(opt->cur==0)
		opt->cur=VEND_OFFSET;
	bytes=len;	
	if(opt->cur+6+bytes>opt->len){
		cp[opt->cur++]=TAG_DHCP_OVL;
		cp[opt->cur++]=1;
		cp[opt->cur++]=1; /* file overload */
		cp[opt->cur]=TAG_END;
		opt->cur=FILE_OFFSET;
	} else if (opt->cur+6+bytes>SNAME_OFFSET && opt->cur<SNAME_OFFSET) {
		cp[opt->cur]=TAG_END;
		for(i=4;(u_long)bp->bp_vend[i]!=TAG_DHCP_OVL;i++);
		bp->bp_vend[i+2]=3;
		opt->cur=SNAME_OFFSET;
	} else if(opt->cur<VEND_OFFSET && opt->cur+6+bytes>VEND_OFFSET)
		return -1;
	cp[opt->cur++]=tag;
	cp[opt->cur++]=bytes;
	bcopy(value,cp+opt->cur,bytes);
	opt->cur+=bytes;

	return 0;
}

int insert_str_opt(bp,opt,tag,value)
struct bootp *bp;
struct opts *opt;
unsigned char tag;
char *value;
{
	return insert_string_opt(bp,opt,tag,value,strlen(value));
}

/* smalloc()  --  safe malloc()
 *
 * Always returns a valid pointer (if it returns at all).  The allocated
 * memory is initialized to all zeros.  If malloc() returns an error, a
 * message is printed using the report() function and the program aborts
 * with a status of 1.
 */

void *
smalloc(nbytes)
	unsigned nbytes;
{
	void *retvalue;

	retvalue = malloc(nbytes);
	if (!retvalue) {
		report(LOG_ERR, "malloc() failure -- exiting");
		exit(1);
	}
	bzero((char *) retvalue, nbytes);
	return retvalue;
}
